from pyProfile.psd import plot_sig_psd, normPSD
import matplotlib.pyplot as plt
import numpy as np
import os
from dataIO.span import span
from dataIO.fn_add_subfix import fn_add_subfix
from dataIO.dicts import strip_kw
# from pyProfile.profile import line
from pyProfile.psd import plot_psd
from pyProfile.psd import psd_units
from pySurf.data2D import plot_data, projection
from pySurf.outliers2d import remove_outliers2d
from matplotlib.colors import LogNorm
from utilities.imaging import fitting as fit
# from IPython.display import display
from plotting.fignumber import fignumber
from plotting.backends import maximize
from pySurf.data2D import levellegendre, level_by_line
from pySurf.data2D import get_data

import pdb


def psd2d(data, x, y, wfun=None, norm=1, rmsnorm=False):
        """Calculate the 2d psd. return freq and psd.
        
        doesnt work with nan.
        use 2d function for psd np.fft.rfft2 for efficiency and mimics
            what done in pySurf.psd.psd
        norm defines the normalization, see function psd.normPSD.
        2017/01/11 broken interface from (x,y,data..),
        added check to correct on the base of sizes."""
        #2017/08/01 complete refactoring, this was internal _psd2d,
        # now is made analogous of 1d pySurf.psd.psd
        # The previous psd2d was including a lot of plotting and output,
        #    it was renamed in plot_psd2d

        #attention, normalization doesn't really work with
        # window, the rms is the rms of the windowed function.
        
        #assert data.shape[0]==data.shape[1]
        if wfun is None:
            win=np.ones(data.shape[0])[:,None]
        else:
            win=wfun(data.shape[0])[:,None]
        N=data.shape[0]
        L=span(y,True)
        yfft=np.fft.rfft2(data*win,axes=[0])

        normfactor=normPSD(N,L,form=norm)
        if rmsnorm:  #normalize to rms of non windowed function
            normfactor=normfactor*np.nanstd(data,axis=0)**2/np.nanstd(data*win,axis=0)**2

        psd=2*normfactor*(np.abs(yfft))**2
        psd[0,:]=psd[0,:]/2.

        freqs = np.fft.rfftfreq(N,np.float(L)/(N-1))

        return freqs,psd

#COMPUTATION
def avgpsd2d(psddata,axis=1,span=False,expand=False):

    """ wrapper for backward compatibilty and tutorial,
    use directly pySurf.data2D.projection() with axis=1 to get same effect.
    return psd average along axis from 2d psd computed with same
    orientation. default axis is 1, because axial profiles
    are along axis 0, sum on axis 1.
    """
    from pySurf.data2D import projection
    return projection(psddata,axis=axis,span=span,expand=expand)

def rms_power(f,p,rmsrange=None,squeeze=True):
    """integrate `p` (2dpsd) to calculate rms slice power in a range of freq. f is the  frequency axis (vertically oriented in plots) for p. Accepts None as extreme of rmsrange. frequencies are assumed to be equally spaced.
    Return a vectors rms with one element for each range in rmsrange. If one element, extra dimension is
        removed unless squeeze is set (useful e.g. to call from wrapper function and get consistent behavior).
    Note that total rms is calculated as rms of column rms, calculated from PSD for each column. If f[0] is zero, the component is excluded, unless first component of rmsrange is explicitly set to zero. If you want to include all frequencies but zero, set rms range first component to None (or rmsrange itself to None).
    Values can then differ from surface rms in case of invalid points (e.g. as consequence of the fact
    that all lines are weighted equally in line average and also invalid points are excluded).
    2020/07/21 corrected handling of zero frequency and added `includezerofreq` Flag. 
    """
    
    #20180404 renamed _rms_power -> rms_power and rms_power -> plot_rms_power
    if rmsrange is None:
        rmsrange=[None,None]
    if rmsrange[0] is None:
        if f[0] == 0:
            rmsrange[0]=f[1]
        else:
            rmsrange[0]=np.min(f,axis=0)
    if rmsrange[1] is None:
        rmsrange[1]=np.max(f,axis=0)

    mask=(f>=rmsrange[0]) & (f<=rmsrange[1])
    df=f[1]-f[0]

    rms=np.sqrt(np.nansum(p[mask,:],axis=0)*df)  #integrate psd for rms
    rms[np.all(np.isnan(p[mask,:]),axis=0)]=np.nan  #set nan where invalid
    #pdb.set_trace()
    if squeeze:
        if len(rms.shape)==1:
            rms.flatten()

    return rms

def plot_rms_power(f,p,x=None,rmsrange=None,ax2f=None,units=None,*args,**kwargs):
    """Plot curves of slice rms from integration of PSD 2D over one or more ranges of frequency `rmsrange`.
    units of x,y,z (scalar or 3-element) can be provided for axis label (as unit of length, not of f,psd),
        None can be used to exclude units. Units of y are irrelevant because integrate, but kept as 3 vector for
        consistency with other routines.
    ax2f can be an array with same length of rmsrange set to True for frequency ranges to be plotted on 2nd
        y (right) axis. Total rms is always plotted on left axis.
    Return a set of slice rms array (of size x), however if rmsrange is not passed returns only a single array. Zero freq is excluded.
    2020/07/21 corrected bug on wrong rms total calculation in title. 
    """
    #20180404 renamed _rms_power -> rms_power and rms_power -> plot_rms_power
    #added units
    if np.size(units)==1:
        units=np.repeat(units,3)

    if rmsrange is not None:
        if len(rmsrange)==2:
            if np.size(rmsrange[0])==1:  #create a nested list if only a value for range
                rmsrange=[rmsrange]

    loc1,loc2=(2 if np.any(ax2f) else 0),1  #these are the position for legends, set first to auto if only one needed, assigned as plt.legend(loc=loc)
    if x is None:
        x=np.arange(p.shape[1])
        
    if f[0] == 0:
        f=f[1:]
        p=p[1:,...]
        
    #set label if not passed in kwargs:
    l='full freq. range [%4.2g : %4.2g]'%(f[0],f[-1])+((" "+units[1] if units[1] is not None else " [Y]")+"$^{-1}$")
    kwargs['label'] = kwargs.pop('label',l)
    
    #calculate and plot rms
    rms=rms_power(f,p,rmsrange=None)
    plt.plot(x,rms,*args,**kwargs)
    
    
    ax3=plt.gca()
    tit1,tit2=(['Left y axis','Right y axis'] if np.any(ax2f) else [None,None]) #legend title headers
    l1=ax3.legend(loc=loc1,title=tit1)

    #plt.title('Total rms power=%6.3g'%(np.sqrt((rms**2).sum()))+((" "+units[2]) if units[2] is not None else ""))  #wrong math, forgets average?
    plt.title('Total rms power=%6.3g'%(np.sqrt(np.nansum(rms**2)/(np.sum(~np.isnan(rms)))))+((" "+units[2]) if units[2] is not None else ""))
    c=plt.gca()._get_lines.prop_cycler

    rms_v=rms
    if rmsrange is not None:
        rms_v=[rms_v]
        if ax2f is None:
            ax2f=np.zeros(len(rmsrange))
        elif np.size(ax2f)==1:   #any sense in having a scalar input?
            ax2f=[ax2f]*len(rmsrange)
        if np.any(ax2f):
            ax4=ax3.twinx()

        #plt.title(plt.title()+' rms=')
        for fr,a in zip(rmsrange,ax2f):
            ax,tit,loc=((ax4,tit2,loc2) if a else (ax3,tit1,loc1))
            #pdb.set_trace()
            rms=rms_power(f,p,rmsrange=fr,squeeze=False)
            if fr[0] is None: fr[0]=min(f)
            if fr[1] is None: fr[1]=max(f)
            sty=next(c)
            l='freq. range [%4.2g:%4.2g]'%(fr[0],fr[1])
            #print(l)
            ax.plot(x,rms,label=l,**sty)
            plt.sca(ax3)
            #plt.title(ax3.get_title()+' [%4.2g : %4.2g]:%4.2g '%(fr[0],fr[1],np.sqrt((rms**2).sum())))
            plt.title(ax3.get_title()+' [%4.2g : %4.2g]:%4.2g '%(fr[0],fr[1],np.nanmean(rms)))
            rms_v.append(rms)
            ax.legend(loc=loc,title=tit)

    ax3.set_xlabel('X'+ (" ("+units[0]+")" if units[0] is not None else ""))
    ax3.set_ylabel('slice rms'+( ' ('+units[2]+')' if units[2] is not None else ' [units of Z]'))
    #ax3.add_artist(l1)
    return rms_v



#PROCESS MULTIPLE DATA AT ONCE
def multipsd2(datalist,wfun=None):
    """return two lists of vectors respectively freq and avg psd2
    each list has vectors respectively for y and x for data must be passed as list of tuples, each one in form (wdata,xwg,ywg)"""
    flist=[]
    psdlist=[]
    for i,(wdata,xwg,ywg) in enumerate(datalist):
        fw1y,pw1y=psd2d(wdata,xwg,ywg) #psds along y
        fw1x,pw1x=psd2d(wdata.T,ywg,xwg) #psds along x
        pw1yavg=avgpsd2d(pw1y)  #y
        pw1xavg=avgpsd2d(pw1x)  #x
        flist.extend([fw1y,fw1x])
        psdlist.extend([pw1yavg,pw1xavg])
        #result.append(((fw1,pw1avg),(fw1m,pw1mavg)))

    return flist,psdlist

def multipsd3(wdata,xwg,ywg,wfun=None):
    """return two lists respectively freq and psd2d
    each list has vectors respectively for y and x for each data set passed.
    data must be passed as list of tuples, each one in form (wdata,xwg,ywg)"""

    fw1y,pw1y=psd2d(wdata,xwg,ywg) #psds along y
    fw1x,pw1x=psd2d(wdata.T,ywg,xwg) #psds along x

    return [fw1y,fw1x],[pw1y,pw1x]

def multipsd(datalist):
    """return a tuple of two couples ((yfreq.,ypsd),(xfreq.,xpsd))
    for each of the data passed, respectively for y and x.
    data must be passed as list of tuples (wdata,xwg,ywg)"""
    result=[]
    for (wdata,xwg,ywg) in datalist:
        fw1,pw1=psd2d(xwg,ywg,wdata) #psds along y
        fw1m,pw1m=psd2d(ywg,xwg,wdata.T) #psds along x
        pw1avg=avgpsd2d(pw1)  #y
        pw1mavg=avgpsd2d(pw1m)  #x
        result.append(((fw1,pw1avg),(fw1m,pw1mavg)))
    return result
    
def plot_psd2d(f,p,x,prange=None,includezerofreq=False,units=None,*args,**kwargs):
    """plots a 2d psd as a surface with logaritmic color scale on the current axis. Return axis.
    If a zero frequency is present it is excluded by default from plot, unless includezerofreq is set to True.
    Units (3-el array) is units of lengths for data (not of PSD), can be None.
    2020/07/10 uncommented call to `pySurf.data2D.plot_data` (in "functions" module `data2D`).
    The "old" code segment called instead directly `plt.imshow`.  
    In some sense, replicates code in `plot_data` and was not necessarily in sync
    (e.g.: stats flag is not included here). 
    This seems to fix the issue. """
    #2018/04/05 critical renaming of plot_psd2d(wdata,x,y..) -> psd2d_analysis
    # and _plot_psd2d(f,p,x) -> plot_psd2d consistently with other routines.

    if len(f.shape)==2:
        print ("""Wrong size detected in call to plot_psd2d. Functions were renamed,
                you probably want to call psd2d_analysis(wdata,x,y..).
                I will try to correct automatically for this time, but you need to
                update calling code URGENTLY, it will be discontinued.""")
        return psd2d_analysis(f,p,x,*args,**kwargs)

    if np.size(units)==1:
        units=np.repeat(units,3)
    #pdb.set_trace()
    #plt.yscale('log')
    plt.title('2D PSD')
    if not includezerofreq:
        if f[0]==0:
            f=f[1:]
            p=p[1:,...]

    if prange is None:
        prange=[None,None]

    # new:
    
    cbunits = psd_units(units)
    #pdb.set_trace()
    ax = plot_data(p,x,f,norm=LogNorm(vmin=prange[0],vmax=prange[1]),
    units=cbunits,aspect='auto',*args,**kwargs)
    plt.ylabel('freq. ('+cbunits[1]+')')
    #pdb.set_trace()
    cb =  plt.gca().images[-1].colorbar
    
    #after introducing psd_units, cbunits is always defined,
    # conditions can be modified, but leaave like this for generality.
    #cb.ax.set_title("PSD"+(" ("+cbunits[2]+")") if cbunits[2] else "")
    
    #plt.ylabel('Frequency ('+cbunits[1]+")")
    #plt.xlabel('X'+ (" ("+cbunits[0]+")" if cbunits[0] else ""))
    plt.grid(1)

    return ax

#PLOTTING

def psd2d_analysis(wdata,x,y,title=None,wfun=None,vrange=None,
    rmsrange=None,prange=None,fignum=5,rmsthr=None, 
    aspect='auto', ax2f=None, units=None,outname=None,norm=1,rmsnorm=True):
    """ Calculates 2D PSD as image obtained combining all profile PSDS calculated along vertical slices of data. Resulting image has size 
    If title is provided rms slice power is also calculated and plotted on three panels with figure and PSD.
    
    Return PSD as PSD2D object. 

    uses plot_rms_power(f,p,rmsrange=None) to calculate rms power.
    fignum window where to plot, if fignnum is 0 current figure is cleared,
    if None new figure is created. Default to figure 5.
    rmsrange is one or a list of frequency ranges for plotting integrated rms. Can contain None to use max or min.
    If axf2 is set to boolean or array of boolean, plot slice rms for the frequencies
    associated to rmsrange on second axis. rmsrange must be set accordingly (same number of elements).
    
    rmsthr sets a threshold for data inclusion. If rms is above the value, the line is considered to contain invalid data and is removed from output PSD.
    This makes it easy to average the returned PSDs.
    Corresponding data are still visualized in central panel, but are marked with a red cross at top of y axes.
    If multiple rms range intervals are provided, line is removed if any of them is over the threshold, but this might change in future, e.g. TODO: make it possible to add vector threshold with as many elements as rms range intervals.

    Set outname to empty string to plot without generating output.

    typical values:
    rmsrange=[[None,0.1],[0.1,0.2],[0.2,1],[1,None]]  #list of frequency intervals for rms
    ax2f=[0,1,1,0,0]             #   axis where to plot rms corresponding to intervals in rmsrange: 1 right, 0 left
    wfun  = np.hanning           #   type of windows for fourier transform
    units = ("mm","mm","$\mu$m") #   units of surface data from which PSD is calculated

    vrange_surf=([-0.5,0.5])  #color scale of surface map
    vrange_leg=([-0.05,0.05])   #color scale of legendre removed map
    prange=np.array((1e-8,1.e-1))#np.array((5e-8,1.e-5))  #color scale of 2d psd plot

    fs,ps=psd2d_analysis(levellegendre(y,wdata,2),x,y,outname="",wfun=np.hanning,
    rmsrange=rmsrange,prange=prange,ax2f=ax2f,units=units)
    """
    #2018/04/05
    # -critical renaming of plot_psd2d(wdata,x,y..) -> psd2d_analysis
    # and _plot_psd2d(f,p,x) -> plot_psd2d consistently with other routines.
    # -added title to determine if generating plot, with the intent of replacing outname.
    plt.pause(1)
    #print('cane',fignum) 
    if outname is not None:  #handle backcompatibility
        print ('psd2d_analysis WARNING: `title` replaced `outname` and output figure will be no more generated.'+
            'OUTNAME will be removed in next version, use title and save the plot after returning from routine.')
        if title is not None:
            print('outname should be not used together with title, it will be ignored.')
        else:
            title=outname
    
    if vrange is not None:
        if prange is None:
            print("""WARNING:  older versions of psd2d use vrange as range of psdvalue.
                    In updated version prange is range of psd and vrange is range for surface
                    plots. Update your code.""")

    #check compatibility with old interface
    if len(wdata.shape)==1:
        if len(y.shape)==2:
            print("WARNING: your code calling psd2d uses old call")
            print("psd2d(x,y,data), automaticly adjusted, but")
            print("update IMMEDIATELY your code to call")
            print("psd2d(data,x,y)")
            x,y,wdata=wdata,x,y

    f, p = psd2d(wdata, x, y, wfun=wfun, norm=norm, rmsnorm=rmsnorm)  # calculate PSD 2D
       
    # GENERATE OUTPUT
    if title is not None:

        if np.size(units)==1:
            units=np.repeat(units,3)

        if prange is None:
            if f[0] == 0:
                prange=span(p[1:,:])
            else:
                prange=span(p)
                if prange[0]<1e-20:  #arbitrary small value
                    print("WARNING: low limit detected in prange, can cause problems with log color scale.")
        if vrange is None:
            vrange=span(wdata)

        #plot surface map, who knows why on figure 5       
        fig=fignumber(fignum)
        #pdb.set_trace()
        #
        plt.clf()
        maximize()
        plt.draw()

        plt.suptitle(title)

        ################
        #plot Surface
        ax1=plt.subplot(311)
        plot_data(wdata,x,y,aspect=aspect,vmin=vrange[0],vmax=vrange[1],units=units,title='Surface',stats=True)
        plt.xlabel(None)
        plt.grid(1)

        ################
        #plot 2D PSD
        ax2=plt.subplot(312,sharex=ax1)
        plot_psd2d(f,p,x,prange=prange,units=units)
        if rmsrange is not None:      #plot horizontal lines to mark frequency ranges for rms extraction,
        #rms is tranformed to 2D array
            if len(np.array(rmsrange).shape)==1:  #if only one interval, make it 2D anyway
                rmsrange=[rmsrange]
            rmsrange=np.array(rmsrange)
            for fr in np.array(rmsrange):  #I want fr to be array type, even if I don't care about rmsrange this does the job
                ax2.hlines(fr[fr != None],*((span(x)-x.mean())*1.1+x.mean()))
        ax2.set_xlabel("")
        plt.xlabel(None)
        #plt.subplots_adjust(top=0.85)

        ################
        #plot rms slice
        ax3=plt.subplot(313,sharex=ax1)
        if f[0]==0: #ignore component of 0 frequency in rms calculation if present
            ff,pp=f[1:],p[1:]
        else:
            ff,pp=f,p

        rms=plot_rms_power(ff,pp,x,rmsrange=rmsrange,ax2f=ax2f,units=units)
        #pdb.set_trace()
        mask=np.isfinite(rms) #True if good. it is an array
        if rmsthr is not None:
            mask = mask & (rms  < rmsthr)
            ax3.hlines(rmsthr,*ax2.get_xlim()) #plot markers on rms chart
            ax2.plot(x[~mask],np.repeat(ax2.get_ylim()[1],len(x[~mask])),'rx') #plot markers on psd2d chart
        
        #pdb.set_trace()
        #questa era qui, ma dava errore perche' mask e' lineare
        #mask = np.all(mask,axis =0) # if any of mask is False -> False
        p[:,~mask]=np.nan
        ax3.grid(1)
        #plt.tight_layout(rect=(0, 0.03, 1, 0.95) if title else (0, 0, 1, 1))
        
        # replaced with more evolved interactive resize:
        #  plt.colorbar().remove() #dirty trick to adjust size to other panels
        def resize(event):
            box1 = ax1.get_position()
            box2 = ax2.get_position()
            box3 = ax3.get_position()
            ax2.set_position([box1.x0, box2.y0, box1.width , box2.height])
            ax2.set_adjustable("box",share=True)
            ax3.set_position([box1.x0, box3.y0, box1.width , box3.height])
            ax3.set_adjustable("box",share=True)

        cid = fig.canvas.mpl_connect('draw_event', resize)
        cid2 = fig.canvas.mpl_connect('resize_event', resize)

        resize(None)

        #plt.tight_layout()

        if outname:    #kept for compatibility, will be removed in next version
            plt.savefig(fn_add_subfix(outname,'_2dpsd','.png'))

    return f,p

def psd_analysis(*args,**kwargs):
    """
    Convenience function to plot average psd.
    Wrapper around `psd2d analysis` analysis. Adds a plot of average PSD and return it. 
    
    Accept parameters for psd2d_analysis and plot_psd,
    including `analysis=True`.
    Call it directly to obtain 2d psd.
    Use 
    Use `plot_psd` from `pySurf.psd` to plot. 
    """
    
    #pdb.set_trace()
    tmp = strip_kw(kwargs,psd2d_analysis,title="",exclude=['units'])
    f,p2=psd2d_analysis(*args,**tmp)
    p=projection(p2,axis=1)
    plt.figure()
    #u = kwargs.get('units',None)
    #t = kwargs.get('title',None)
    kw = strip_kw(kwargs,plot_psd)
    #u = kw['units'] if 'units' in kw else None
    
    #pdb.set_trace()
    plot_psd(f,p,**kw)

    #plot_psd(f,p,units=[u[0],u[1]] if u else u, title= t, **kwargs)
    return f,p


def compare_2dpsd(data,ldata,x=None,y=None,fignum=None,titles=None,vmin=None,vmax=None):
    """plot a four panel data and psd for the two figures from data and ldata.
    x and y must be the same for the two data sets.
    Calculate PSD. Return freqs, psd and lpsd."""

    if x is None or y is None:
        raise NotImplementedError
    freqs,p=psd2d(x,y,data)
    freqs,pl=psd2d(x,y,ldata)

    #ldata=remove_outliers2d(data,nsigma=3,fignum=fignum+1)
    if fignum is not None:
        plt.figure(fignum)
        plt.clf()
        ax1=plt.subplot(221)
        if titles is not None:
            plt.title(titles[0])
        plt.xlabel('X (mm)')
        plt.ylabel('Y (mm)')
        axim=plt.imshow(data,extent=(x[0],x[-1],y[0],y[-1]),
                        interpolation='none',aspect='equal')
        plt.colorbar()

        ax2=plt.subplot(222,sharex=ax1)

        plt.title('PSD')

        plt.imshow(p,norm=LogNorm(vmin=vmin,vmax=vmax),interpolation='none',
                   extent=(x[0],x[-1],freqs[0],freqs[-1]),aspect='auto')
        plt.xlabel('X (mm $\mu$m^2)')
        plt.ylabel('Freq (mm^-1)')
        plt.colorbar()
        plt.draw()

        ax3=plt.subplot(223,sharex=ax1,sharey=ax1)
        if titles is not None:
                    plt.title(titles[1])
        plt.xlabel('X (mm $\mu$m^2)')
        plt.ylabel('Y (mm)')
        axim=plt.imshow(ldata,extent=(x[0],x[-1],y[0],y[-1]),interpolation='none',aspect='equal')
        plt.colorbar()

        plt.subplot(224,sharex=ax1)

        plt.title('PSD outliers corrected')
        plt.imshow(pl,norm=LogNorm(vmin=vmin,vmax=vmax),interpolation='none',
                   extent=(x[0],x[-1],freqs[0],freqs[-1]),aspect='auto')
        plt.xlabel('X (mm)')
        plt.ylabel('Yfreq (mm^-1)')
        plt.colorbar()
        plt.draw()

    return freqs,p,pl

#SCRIPTS
def avgpsd(fitsfile,correct=False,reader=None,level=None,nsigma=3,rotate=False,**kwargs):
    """Wrapper on avgpsd2d. Extract average psd from a fits image,
    includes reading and outliers removal.
    Reader is a function of fitsfile that returns x,y,data."""
    if reader is None:  #I don't care it's not pythonic, simplest way to avoid circular imports.
        from pySurf.readers.instrumentReader import getdata
        reader=getdata

    x,y,data=reader(fitsfile) #x and y in mm, data in um
    if rotate:
        data=data.T
        x,y=y,x
    if level is not None:
        data=level_by_line(data,**kwargs)
    if correct:
        data=remove_outliers2d(data,nsigma=nsigma)
    f,p=psd2d(x,y,data)
    pm=avgpsd2d(p,axis=1)
    return f,pm

def calculatePSD(wdata,xg,yg,outname="",wfun=None,vrange=None,rmsrange=None,prange=None,fignum=1):
    """given points w, calculate and plot surface maps with different leveling (piston, tilt, sag, 10 legendre)
    use psd2d to calculate and save x and y 2d PSDs, plots only y.
    fignum window where to plot, if fignnum is 0 current figure is cleared, if None new figure is created. Default to figure 1 ."""
    ###AWFUL CODE, partially fixed -> fix plotting window

    ## FOUR PANEL PLOT save as _lev.png, _piston.dat, _tilt.dat, _sag.dat
    ## on fignum
    if fignum==0:
        plt.clf()
    else:
        plt.figure(fignum)
    plt.clf()
    figManager = plt.get_current_fig_manager()
    figManager.window.showMaximized()
    #xg,yg=points_find_grid(w,'grid')[1]
    wdata,lwdata,lwdata2=leveldata(wdata,xg,yg) #calculate and make a plot with 3 panels
    if outname:
        np.savetxt(fn_add_subfix(outname,'_piston','.dat'),wdata) #doesn't  write x and y
        np.savetxt(fn_add_subfix(outname,'_tilt','.dat'),lwdata)
        np.savetxt(fn_add_subfix(outname,'_sag','.dat'),lwdata2)

    leg10=wdata-fit.legendre2d(wdata,10,10)[0]
    #forth panel contains legendre
    plt.subplot(224)
    plt.imshow(leg10,extent=(xg[0],xg[-1],yg[0],yg[-1]),
                interpolation='None',aspect='auto')
    plt.title('10 legendre')
    plt.colorbar()
    #display(plt.gcf())
    if outname:
        plt.savefig(fn_add_subfix(outname,'_lev','.png'))

    ## DATA OUTPUT save as _psd.dat, psd2d creates
    #create packed data for txt output
    #containing x and y psds, with 3 different levelings
    #resulting in 6 couples freq/psd.
    # matches the order specified in header

    #header for final matrix output, define labels of psd sequence
    head='yfreq ypsd xfreq xpsd ylevfreq ylevpsd xlevfreq xlevpsd ysagfreq ysagpsd xsagfreq xsagpsd'

    #this assumes x and y are same xg and yg on all datalist
    datalist=[wdata,lwdata,lwdata2]
    labels=['_piston','_tilt','_sag'] #postfix for output file
    flist=[]
    psdlist=[]
    for d,l in zip(datalist,labels):
        plt.figure(3)
        #x and y psds
        fx,px=psd2d(d.T,yg,xg,wfun=wfun) #psds along x, no plot
        fy,py=plot_psd2d(d,xg,yg,outname=fn_add_subfix(outname,l),wfun=wfun,
            rmsrange=rmsrange,prange=prange,vrange=vrange) #psds along y, plot
        #display(plt.gcf())

        #note y goes first
        flist.extend([fy,fx])
        psdlist.extend([py,px])
    avgpsdlist=[avgpsd2d(p) for p in psdlist]

    #generate output array and save it with header
    outarr=np.empty((np.max([len(f) for f in flist]),len(psdlist)*2))*np.nan
    for i,(f,p) in enumerate(zip(flist,avgpsdlist)):
        outarr[:len(f),i*2]=f
        outarr[:len(p),i*2+1]=p

    if outname:
        np.savetxt(fn_add_subfix(outname,'_psd','.dat'),outarr,header=head,fmt='%f')

    ## PLOT AVG PSDs save as _psds.png
    fw1,fw1m,fw2,fw2m,fw3,fw3m=flist  #used only in plot
    pw1avg,pw1mavg,pw2avg,pw2mavg,pw3avg,pw3mavg=avgpsdlist  #average psds
    plt.figure(2)
    plt.clf()
    plt.plot(fw1,pw1avg,label='Y')
    plt.plot(fw2,pw2avg,label='Y lev.')
    plt.plot(fw3,pw3avg,label='Y sag.')
    plt.plot(fw1m,pw1mavg,'--',label='X')
    plt.plot(fw2m,pw2mavg,'--',label='X lev.')
    plt.plot(fw3m,pw3mavg,'--',label='X sag.')
    plt.loglog()
    plt.grid()
    plt.legend(loc=0)
    if outname:
        plt.savefig(fn_add_subfix(outname,'_psds','.png'))

    return outarr


def calculatePSD2(wdata,xg,yg,outname="",wfun=None,vrange=[None,None],rmsrange=None,prange=None,fignum=1,misal_deg=(1,1),leg_deg=(10,10)):
    """Updated version with subtracted 4 terms legendre."""
    """given points w, calculate and plot surface maps with different leveling (piston, tilt, sag, 10 legendre)
    use psd2d to calculate and save x and y 2d PSDs, plots only y.
    fignum window where to plot, if fignnum is 0 current figure is cleared, if None new figure is created. Default to figure 1 ."""
    
    from plotting.multiplots import compare_images
    ##STILL AWFUL CODE
    # misal_deg is useless, since data are expected to be already leveled.
    # if not, still can be applied, but is quite useless.

    ###AWFUL CODE, partially fixed -> fix plotting window

    ## FOUR PANEL PLOT save as _lev.png, _piston.dat, _tilt.dat, _sag.dat
    ## on fignum
    fig=fignumber(fignum)
    plt.clf()
    figManager = plt.get_current_fig_manager()
    figManager.window.showMaximized()

    #
    wdata=-fit.legendre2d(wdata,1,1)[0]
    lwdata=wdata-fit.legendre2d(wdata,*misal_deg)[0]
    lwdata2=wdata-fit.legendre2d(wdata,*leg_deg)[0]
    #
    #make plots
    ax=compare_images([wdata,lwdata,lwdata2],xg,yg,titles=['original','cone corrected','10 legendre 2D corrected'],fignum=0,commonscale=True,vmin=vrange[0],vmax=vrange[1])  #generator of axis, not sure how it can work below, missing something?
    for im in ax:
        plt.xlabel('X (mm)')
        plt.ylabel('Y (mm)')
    #
    ##return wdata,lwdata,lwdata2

    wdata,lwdata,lwdata2=leveldata(wdata,xg,yg) #calculate and make a plot with 3 panels
    if outname:
        np.savetxt(fn_add_subfix(outname,'_piston','.dat'),wdata) #doesn't  write x and y
        np.savetxt(fn_add_subfix(outname,'_%2i_%2i_misal'%misal_deg,'.dat'),lwdata)
        np.savetxt(fn_add_subfix(outname,'_%2i_%2i_leg'%leg_deg,'.dat'),lwdata2)

    leg10=wdata-fit.legendre2d(wdata,*leg_deg)[0]
    #forth panel contains legendre
    plt.subplot(224)
    plt.imshow(leg10,extent=(xg[0],xg[-1],yg[0],yg[-1]),
                interpolation='None',aspect='auto')
    plt.title('(%i,%i) 2D legendre removed'%leg_deg)
    plt.colorbar()
    #display(plt.gcf())
    if outname:
        plt.savefig(fn_add_subfix(outname,'_lev','.png'))

    ## DATA OUTPUT save as _psd.dat, psd2d creates
    #create packed data for txt output
    #containing x and y psds, with 3 different levelings
    #resulting in 6 couples freq/psd.
    # matches the order specified in header

    #header for final matrix output, define labels of psd sequence
    head='yfreq ypsd xfreq xpsd ylevfreq ylevpsd xlevfreq xlevpsd ysagfreq ysagpsd xsagfreq xsagpsd'

    #this assumes x and y are same xg and yg on all datalist
    #change this part, we are no more interested in doing psd on all possible leveling, only 4 terms removed by line
    # along the direction of interest, however to keep constant number of columns, replace piston, tilt, sag
    # with tilt, sag, 4 legendre
    datalist=[wdata,lwdata,lwdata2]
    labels=['_tilt','_mis','_leg'] #postfix for output file
    flist=[]
    psdlist=[]
    for d,l in zip(datalist,labels):
        plt.figure(3)
        #x and y psds
        fx,px=psd2d(d.T,yg,xg,wfun=wfun) #psds along x, no plot
        fy,py=plot_psd2d(d,xg,yg,outname=fn_add_subfix(outname,l),wfun=wfun,
            rmsrange=rmsrange,prange=prange,vrange=vrange) #psds along y, plot
        #display(plt.gcf())

        #note y goes first
        flist.extend([fy,fx])
        psdlist.extend([py,px])
    avgpsdlist=[avgpsd2d(p) for p in psdlist]

    #generate output array and save it with header
    outarr=np.empty((np.max([len(f) for f in flist]),len(psdlist)*2))*np.nan
    for i,(f,p) in enumerate(zip(flist,avgpsdlist)):
        outarr[:len(f),i*2]=f
        outarr[:len(p),i*2+1]=p

    if outname:
        np.savetxt(fn_add_subfix(outname,'_psd','.dat'),outarr,header=head,fmt='%f')

    ## PLOT AVG PSDs save as _psds.png
    fw1,fw1m,fw2,fw2m,fw3,fw3m=flist  #used only in plot
    pw1avg,pw1mavg,pw2avg,pw2mavg,pw3avg,pw3mavg=avgpsdlist  #average psds
    plt.figure(2)
    plt.clf()
    plt.plot(fw1,pw1avg,label='Y')
    plt.plot(fw2,pw2avg,label='Y lev.')
    plt.plot(fw3,pw3avg,label='Y sag.')
    plt.plot(fw1m,pw1mavg,'--',label='X')
    plt.plot(fw2m,pw2mavg,'--',label='X lev.')
    plt.plot(fw3m,pw3mavg,'--',label='X sag.')
    plt.loglog()
    plt.grid()
    plt.legend(loc=0)
    if outname:
        plt.savefig(fn_add_subfix(outname,'_psds','.png'))

    return outarr


'''
def plot_profile_outliers(data,x,y,ldata,i=1000):
    print("obsolete, this routine will be removed from psd2d")
    plt.figure(2)
    plt.clf()
    plot_sig_psd(x,data[:,i],scale=1,label='data')
    plot_sig_psd(x,ldata[:,i],scale=1.,label='outliers removed')
    plt.suptitle('Effects of outliers removal - %s'%datafile).set_size('large')

    plt.grid()
    plt.legend(loc=0)
    plt.subplot(211)
    plt.grid()

    plt.subplot(212)
    freqs,p=psd2d(x,y,data)
    pm=avgpsd2d(p,axis=1)
    plt.plot(freqs,pm,label='2D average raw')
    lpm=avgpsd2d(psd2d(x,y,ldata)[1],axis=1)
    plt.plot(freqs,lpm,label='2D average corr.')
    plt.ylabel('(mm $\mu$m^2)')
    plt.xlabel('mm^-1')
    plt.legend(loc=0)
'''

#####################################################

##TESTS

def test_slicerms():
    """test plot options for slicerms """
    wf=r'test\PSD\2dpsd\171010_Secondary_Mandrel_C3_RefSub.dat'
    outfolder=r'test\PSD\2dpsd\output'
    wdata,x,y=get_data(wf,matrix=True,addaxis=True)
    outfile=fn_add_subfix(wf,'_4leg',strip=True,pre=outfolder+os.path.sep)

    f3,psd3=psd2d(levellegendre(y,wdata,4),x,y,wfun=np.hanning,norm=1,rmsnorm=True)
    plt.figure()
    r=plot_rms_power(f3,psd3,x=None,rmsrange=None)
    plt.savefig(fn_add_subfix(outfile,'_testslicerms_01','.png'))
    plt.figure()
    plot_rms_power(f3,psd3,x=x,rmsrange=[0.1,1])
    plt.savefig(fn_add_subfix(outfile,'_testslicerms_02','.png'))
    plt.figure()
    plot_rms_power(f3,psd3,x=x,rmsrange=[0.1,1],ax2f=[1])
    plt.savefig(fn_add_subfix(outfile,'_testslicerms_03','.png'))
    plt.figure()
    plot_rms_power(f3,psd3,x=x,rmsrange=[[None,0.1],[0.1,1],[1,None]],ax2f=[0,1,1])
    plt.savefig(fn_add_subfix(outfile,'_testslicerms_04','.png'))

def test_plotpsd2d():
    """test of 3 panel psd2d plot, including 2d psd and slice rms. """
    prange=np.array((1e-10,1))#np.array((5e-8,1.e-5))  #color scale of 2d psd plot
    rmsrange=[[0.1,1],[0.01,0.1]]  #range of frequency for rms calculation
    vrange_surf=([-0.15,0.15])  #color scale of surface map
    vrange_leg=([-0.025,0.025])   #color scale of legendre removed map

    wf=r'test\PSD\2dpsd\171010_Secondary_Mandrel_C3_RefSub.dat'
    outfolder=r'test\PSD\2dpsd\output'
    wdata,x,y=get_data(wf,matrix=True,addaxis=True)
    outfile=fn_add_subfix(wf,'_4leg',strip=True,pre=outfolder+os.path.sep)

    plt.figure()
    #PSD with polynomial removal and average in f3, psd3 and wptot
    f3,psd3=plot_psd2d(levellegendre(y,wdata,2),x,y,wfun=np.hanning,
                outname=outfile,vrange=vrange_leg,prange=prange,rmsrange=rmsrange,ax2f=True)

def test_psd2d():
    """test of plotting 2dpsd alone """
    from pySurf.data2D import get_data, levellegendre

    rmsrange=[[None,0.1],[0.1,0.2],[0.2,1],[1,None]]  #list of frequency intervals for rms
    ax2f=[0,1,1,0,0]             #   axis where to plot rms corresponding to intervals in rmsrange: 1 right, 0 left
    wfun  = np.hanning           #   type of windows for fourier transform
    units = ("mm","mm","$\mu$m") #   units of surface data from which PSD is calculated

    vrange_surf=([-0.5,0.5])  #color scale of surface map
    vrange_leg=([-0.05,0.05])   #color scale of legendre removed map
    prange=np.array((1e-8,1.e-3))#np.array((5e-8,1.e-5))  #color scale of 2d psd plot
    order=2                     #highest legendre order to remove (2=sag)

    TEST_PATH=os.path.join(os.path.dirname(os.path.realpath(__file__)),r'test\PSD\2dpsd')
    wf=os.path.join(TEST_PATH,r'171010_Secondary_Mandrel_C3_RefSub.dat')
    OUTFOLDER=os.path.join(TEST_PATH,'output')
    OUTFILE=fn_add_subfix(wf,'_4leg',strip=True,pre=OUTFOLDER+os.path.sep)

    wdata,x,y=get_data(wf,matrix=True,addaxis=True)
    fs,ps=psd2d_analysis(levellegendre(y,wdata,order),x,y,outname=OUTFILE,wfun=np.hanning,
        rmsrange=rmsrange,prange=prange,ax2f=ax2f,units=units)

    f3,psd3=psd2d(levellegendre(y,wdata,order),x,y,wfun=np.hanning,norm=1,rmsnorm=True)

    plt.figure(1)
    plt.clf()
    plot_psd2d(f3,psd3,x,rmsrange=rmsrange)
    plt.figure(2)
    plt.clf()
    plot_psd2d(f3,psd3,x,rmsrange=rmsrange,prange=prange)

    """
    plt.figure()
    plt.title('2D PSD')
    plt.imshow(psd3,norm=LogNorm(),interpolation='none',
        extent=(x[0],x[-1],f3[0],f3[-1]),origin='lower',aspect='auto')
    plt.colorbar()
    """

    plt.figure(3)
    plt.clf()
    plt.title('2D PSD')
    plt.imshow(psd3,norm=LogNorm(vmin=prange[0],vmax=prange[1]),interpolation='none',
        extent=(x[0],x[-1],f3[0],f3[-1]),origin='lower',aspect='auto')
    plt.colorbar()


def mwc(psd3,prange):
    """plots a test case"""
    plt.title(str(prange))
    plt.imshow(psd3,norm=LogNorm(vmin=prange[0],vmax=prange[1]),interpolation='none',
        origin='lower',aspect='equal')
    from matplotlib.ticker import LogFormatter
    formatter = LogFormatter(10, labelOnlyBase=False)
    plt.colorbar(format=formatter)  #ticks=[1,5,10,20,50],

    #array([  5.03059356e-13,   4.11978017e-01])

def test_prange_mwc():
    from plotting.multiplots import compare_images,subplot_grid

    ptest=[(1e-9,0.1),(1e-11,0.1),(1e-13,0.1)]
    datarange=np.array([  5.0e-13,   4e-01])
    random_sample=np.random.random_sample
    psd3=random_sample((100,50))*(datarange[1]-datarange[0])+datarange[0]
    plt.figure()

    for i,p in enumerate(ptest):
        plt.subplot (1, 3, i+1)
        mwc(psd3,p)
    #compare_images(*zip(psd3,x,p),fignum=0,titles=map(str,ptest))
    plt.tight_layout()

def test_prange_mwc2():
    from plotting.multiplots import compare_images,subplot_grid

    #make data
    datarange=np.array([  5.0e-13,   4e-01])
    random_sample=np.random.random_sample
    psd3=random_sample((100,50))*(datarange[1]-datarange[0])+datarange[0]

    ptest=[(1e-9,0.1),(1e-11,0.1),(1e-13,0.1)] #test values for color range
    plt.figure()
    for i,p in enumerate(ptest):
        plt.subplot (1, 3, i+1)
        plt.title(str(p))
        plt.imshow(psd3,norm=LogNorm(vmin=p[0],vmax=p[1]))
        plt.colorbar()
    plt.tight_layout()



#This submodule facilitates stitching of metrology images
#The images must have fiducials to compute translation and rotation
#Piston/tip/tilt is done by arrays after translation/rotation is fixed
#by fiducials
import utilities.transformations as tr
import numpy as np
import matplotlib.pyplot as plt
from . import man
from scipy.optimize import minimize
from .analysis import getPoints
from scipy.interpolate import griddata
from utilities.plotting import nanmean

import pdb

def transformCoords(x,y,tx,ty,theta):
    """Transforms coordinates x,y by translating tx,ty
    and rotation theta about x
    Returns: x,y of new coords
    """
    trans = tr.translation_matrix([tx,ty,0])
    rot = tr.rotation_matrix(theta,[0,0,1],point=[np.mean(x),np.mean(y),0])
    pos0 = np.array((x,y,np.repeat(0.,np.size(x)),np.repeat(1.,np.size(x))))
    pos1 = np.dot(trans,np.dot(rot,pos0))

    return pos1[0],pos1[1]

def transformCoords_wMag(x,y,tx,ty,theta,mag):
    """Transforms coordinates x,y by magnifying x,y by a constant factor,
    then translating tx,ty and rotating theta about x
    Returns: x,y of new coords
    """
    mag_x,mag_y = x*mag,y*mag
    trans = tr.translation_matrix([tx,ty,0])
    rot = tr.rotation_matrix(theta,[0,0,1],point=[np.mean(x),np.mean(y),0])
    pos0 = np.array((mag_x,mag_y,np.repeat(0.,np.size(x)),np.repeat(1.,np.size(x))))
    pos1 = np.dot(trans,np.dot(rot,pos0))
    return pos1[0],pos1[1]

def transformCoords_wSeparateMag(x,y,tx,ty,theta,x_mag,y_mag):
    """Transforms coordinates x,y by magnifying x,y by a constant factor,
    then translating tx,ty and rotating theta about x
    Returns: x,y of new coords
    """
    new_x,new_y = x*x_mag,y*y_mag
    trans = tr.translation_matrix([tx,ty,0])
    rot = tr.rotation_matrix(theta,[0,0,1],point=[np.mean(x),np.mean(y),0])
    pos0 = np.array((new_x,new_y,np.repeat(0.,np.size(x)),np.repeat(1.,np.size(x))))
    pos1 = np.dot(trans,np.dot(rot,pos0))
    return pos1[0],pos1[1]

def matchFiducials(x1,y1,x2,y2):
    """This function will compute a rotation and translation
    to match a list of fiducial coordinates
    Returns: translation tx,ty and rotation theta about zhat
    to bring x2,y2 to x1,y1
    """
    #Make function to minimize
    fun = lambda p: sumOfSquares(x1,y1,*transformCoords(x2,y2,*p))

    #Make starting guess
    start = zeros(3)
    start[0] = mean(x1-x2)
    start[1] = mean(y1-y2)
    start[2] = .0001

    #Run minimization and return fiducial transformation
    res = minimize(fun,start,method='nelder-mead',\
                   options={'disp':True,'maxfev':1000})
    
    return res['x']

def matchFiducials_wMag(x1,y1,x2,y2):
    '''
    This function will compute a rotation, translation and
    magnification needed to match a list of fiducial coordinates.
    Returns:
    tx, ty - translations
    theta - rotation
    mag - magnification factor
    These transformations are needed to bring x2,y2 to x1,y1
    '''

    #Make function to minimize
    fun = lambda p: sumOfSquares(x1,y1,*transformCoords_wMag(x2,y2,*p))
    
    #Make starting guess
    start = np.zeros(4)
    start[0] = np.mean(x1-x2)
    start[1] = np.mean(y1-y2)
    start[2] = .0001
    start[3] = 1.0

    #Run minimization and return fiducial transformation
    res = minimize(fun,start,method='nelder-mead',\
                   options={'disp':True,'maxfev':1000})
    
    return res['x']

def matchFiducials_wSeparateMag(x1,y1,x2,y2):
    '''
    This function will compute a rotation, translation and
    magnification needed to match a list of fiducial coordinates.
    Returns:
    tx, ty - translations
    theta - rotation
    mag - magnification factor
    These transformations are needed to bring x2,y2 to x1,y1
    '''

    #Make function to minimize
    fun = lambda p: sumOfSquares(x1,y1,*transformCoords_wSeparateMag(x2,y2,*p))
    
    #Make starting guess
    start = np.zeros(5)
    start[0] = np.mean(x1-x2)
    start[1] = np.mean(y1-y2)
    start[2] = .0001
    start[3] = 1.0
    start[4] = 1.0

    #Run minimization and return fiducial transformation
    res = minimize(fun,start,method='nelder-mead',\
                   options={'disp':True,'maxfev':1000})
    
    return res['x']

def sumOfSquares(x1,y1,x2,y2):
    """Computes the sum of the squares of the residuals
    for two lists of coordinates
    """
    return sum(np.sqrt((x1-x2)**2+(y1-y2)**2))

def matchPistonTipTilt(img1,img2):
    """This function applies piston and tip/tilt
    to minimize RMS difference between two arrays
    Returns: img2 matched to img1
    """
    #Make function to minimize
    fun = lambda p: nanmean((img1 - man.tipTiltPiston(img2,*p))**2)

    #Run minimization and return matched image
    res = minimize(fun,[1.,.1,.1],method='nelder-mead',\
                   options={'disp':True,'maxfev':1000})

    return man.tipTiltPiston(img2,*res['x'])

def stitchImages(img1,img2):
    """Allows user to pick fiducials for both images.
    Function then computes the transform to move img2
    to img1 reference frame.
    Updated
    """
    #Get both fiducials
    xf1,yf1 = getPoints(img1)
    xf2,yf2 = getPoints(img2)

    #Match them
    tx,ty,theta = matchFiducials(xf1,yf1,xf2,yf2)

    #Pad img1 based on translations
    img1 = man.padNaN(img1,n=round(tx),axis=1)
    img1 = man.padNaN(img1,n=round(ty),axis=0)
    #Shift img1 fiducial locations
    if tx<0:
        xf1 = xf1 - tx
    if ty<0:
        yf1 = yf1 - ty

    #Get x,y,z points from stitched image
    x2,y2,z2 = man.unpackimage(img2,xlim=[0,shape(img2)[1]],\
                           ylim=[0,shape(img2)[0]])

    #Apply transformations to x,y coords
    x2,y2 = transformCoords(x2,y2,ty,tx,theta)

    #Get x,y,z points from reference image
    x1,y1,z1 = man.unpackimage(img1,remove=False,xlim=[0,shape(img1)[1]],\
                           ylim=[0,shape(img1)[0]])

    #Interpolate stitched image onto expanded image grid
    newimg = griddata((x2,y2),z2,(x1,y1),method='linear')
    print('Interpolation ok')
    newimg = newimg.reshape(shape(img1))

    #Images should now be in the same reference frame
    #Time to apply tip/tilt/piston to minimize RMS
    newimg = matchPistonTipTilt(img1,newimg)

    #Would like list of enlarge image showing all valid data, this is final step
    #Avoid overwritting fiducials
    #Save indices of NaNs near fiducials
    find = logical_and(sqrt((y1-xf1[0])**2+(x1-yf1[0])**2) < 15.,\
                       isnan(img1).flatten())
    for i in range(1,size(xf1)):
        find = logical_or(find,\
                logical_and(sqrt((y1-xf1[i])**2+(x1-yf1[i])**2) < 15.,\
                        isnan(img1).flatten()))

    #Include newimg data
    img1[isnan(img1)] = newimg[isnan(img1)]
    #Reset fiducials to NaNs
    img1[find.reshape(shape(img1))] = NaN

    #Return translations to automatically pad next image
    return img1,tx,ty

def overlapTrans(x,y,tx,ty,theta,sx,sy):
    """
    Transform coordinates with a rotation about mean coordinate,
    followed by translation, followed by scaling
    """
    x2,y2 = x*np.cos(theta)+y*np.sin(theta),\
            -x*np.sin(theta)+y*np.cos(theta)
    x2,y2 = x2+tx, y2+ty
    x2,y2 = x2*sx, y2*sy
    
    return x2,y2

def overlapMerit(x1,y1,z1,x2,y2,z2,tx,ty,theta,sx,sy):
    """
    Apply transformation on img2 and return RMS error
    """
    #Apply coordinate transformation
    x3,y3 = overlapTrans(x2,y2,tx,ty,theta,sx,sy)
    #Interpolate onto img1 grid
    x3,y3,z2 = x3.flatten(),y3.flatten(),z2.flatten()
    ind = ~np.isnan(z2)
    x3,y3,z2 = x3[ind],y3[ind],z2[ind]
    z4 = griddata((x3.flatten(),y3.flatten()),\
                  z2.flatten(),(x1,y1),method='cubic')
    #Determine overlap area
    area = np.sum(np.logical_and(~np.isnan(z1),~np.isnan(z4)))
    #Return error
    resid = z1-z4
    resid = resid - np.nanmean(resid)
    return np.sqrt(np.nanmean(resid**2)),z4

def overlapImages(img1,img2,scale=False):
    """Function to interpolate a second image onto the first.
    This is used to compare metrology carried out on a part
    before and after some processing step. The first image may
    be translated, rotated, and scaled with respect to the first
    image. Scaling is due to magnification changes.
    Procedure is to:
    1) Set NaNs to zeros so that they want to overlap
    2) Set up merit function as function of transformation
       and scaling. Return RMS error.
    3) Use optimizer to determine best transformation
    4) Interpolate img2 to img1 using this transformation
       and return transformed img2.
    """
    #Unpack images
    x1,y1 = np.meshgrid(np.linspace(-1.,1.,np.shape(img1)[1]),\
                        np.linspace(-1.,1.,np.shape(img1)[0]))
    x2,y2 = np.meshgrid(np.linspace(-1.,1.,np.shape(img2)[1]),\
                        np.linspace(-1.,1.,np.shape(img2)[0]))
    #Get centroids of each image
    cx1 = np.mean(x1[~np.isnan(img1)])
    cy1 = np.mean(y1[~np.isnan(img1)])
    cx2 = np.mean(x1[~np.isnan(img2)])
    cy2 = np.mean(y1[~np.isnan(img2)])

    #Save NaN index
    nans = np.isnan(img1)

    #Set up merit function
    if scale is False:
        fun = lambda p: overlapMerit(x1,y1,img1,x2,y2,img2,\
                                     p[0],p[1],p[2],1,1)[0]
        start = [cx1-cx2+.01,cy1-cy2+.01,0.1]
        print(start)
    else:
        fun = lambda p: overlapMerit(x1,y1,img1,x2,y2,img2,\
                                     *p)[0]
        start = [0.1,0.1,.1,1,1]

    #Optimize!
    pdb.set_trace()
    res = minimize(fun,start,method='Powell',\
                   options={'disp':True,'maxfev':1000,\
                            'ftol':.0001,\
                            'xtol':.0001})

    #Construct transformed img2
    imgnew = overlapMerit(x1,y1,img1,x2,y2,img2,\
                          res['x'][0],res['x'][1],res['x'][2],\
                          1,1)[1]
    img1[nans] = np.nan
    imgnew[nans] = np.nan

    pdb.set_trace()

    return imgnew

def AlignImagesWithFiducials(img1,img2,xf1,yf1,xf2,yf2):
    """
    Aligns img2 to img1 based on an array listing the x,y coordinates of common fiducials.
    Arguments:
    img1 - the reference image to be aligned to.
    img2 - the image to be aligned to the reference image.
    xf1 - an array containing the x coordinates of the fiducials in img1
    yf1 - an array containing the y coordinates of the fiducials in img1.
    xf2 - an array containing the x coordinates of the fiducials in img2.
    yf2 - an array containing the y coordinates of the fiducials in img2.
    Returns:
    newimg - img2 as aligned and interpolated to the coordinates of img1.
    """
    #Match them
    tx,ty,theta,mag = matchFiducials_wMag(yf1,xf1,yf2,xf2)

    x2_wNaNs,y2_wNaNs,z2_wNaNs = man.unpackimage(img2,remove = False,xlim=[0,np.shape(img2)[1]],\
                           ylim=[0,np.shape(img2)[0]])
    #Apply transformations to x,y coords
    x2_wNaNs,y2_wNaNs = transformCoords_wMag(x2_wNaNs,y2_wNaNs,ty,tx,theta,mag)
    
    #Get x,y,z points from reference image
    x1,y1,z1 = man.unpackimage(img1,remove=False,xlim=[0,np.shape(img1)[1]],\
                           ylim=[0,np.shape(img1)[0]])

    #Interpolate stitched image onto expanded image grid
    newimg = griddata((x2_wNaNs,y2_wNaNs),z2_wNaNs,(x1,y1),method='linear')
    print('Interpolation ok')
    newimg = newimg.reshape(np.shape(img1))

    #Images should now be in the same reference frame
    #Time to apply tip/tilt/piston to minimize RMS
    newimg = matchPistonTipTilt(img1,newimg)

    return newimg

def AlignImagesWithFiducials_SeparateMag(img1,img2,xf1,yf1,xf2,yf2):
    """
    Aligns img2 to img1 based on an array listing the x,y coordinates of common fiducials.
    Arguments:
    img1 - the reference image to be aligned to.
    img2 - the image to be aligned to the reference image.
    xf1 - an array containing the x coordinates of the fiducials in img1
    yf1 - an array containing the y coordinates of the fiducials in img1.
    xf2 - an array containing the x coordinates of the fiducials in img2.
    yf2 - an array containing the y coordinates of the fiducials in img2.
    Returns:
    newimg - img2 as aligned and interpolated to the coordinates of img1.
    """
    #Match them
    tx,ty,theta,x_mag,y_mag = matchFiducials_wSeparateMag(yf1,xf1,yf2,xf2)

    x2_wNaNs,y2_wNaNs,z2_wNaNs = man.unpackimage(img2,remove = False,xlim=[0,np.shape(img2)[1]],\
                           ylim=[0,np.shape(img2)[0]])
    #Apply transformations to x,y coords
    x2_wNaNs,y2_wNaNs = transformCoords_wSeparateMag(x2_wNaNs,y2_wNaNs,ty,tx,theta,x_mag,y_mag)
    
    #Get x,y,z points from reference image
    x1,y1,z1 = man.unpackimage(img1,remove=False,xlim=[0,np.shape(img1)[1]],\
                           ylim=[0,np.shape(img1)[0]])

    #Interpolate stitched image onto expanded image grid
    newimg = griddata((x2_wNaNs,y2_wNaNs),z2_wNaNs,(x1,y1),method='linear')
    print('Interpolation ok')
    newimg = newimg.reshape(np.shape(img1))

    #Images should now be in the same reference frame
    #Time to apply tip/tilt/piston to minimize RMS
    newimg = matchPistonTipTilt(img1,newimg)

    return newimg


from numpy import *
import numpy as np
from matplotlib.pyplot import *
from matplotlib import ticker,colors
import pickle,pdb
from mpl_toolkits.mplot3d import Axes3D

def scatter3d(x,y,z,**args):
    """Make a 3d scatter plot"""
    fig = figure()
    ax = fig.add_subplot(111,projection='3d')
    ax.scatter(x,y,z)
    ax.set_xlabel('x')
    ax.set_ylabel('y')
    ax.set_zlabel('z')
    return ax

#Make iso plot
def isoplot(*args,**kargs):
    fig = gcf()
    #fig.clf()
    ax = fig.add_subplot(111,aspect='equal')
    ax.plot(*args,**kargs)
    fig.show()
    return ax

#Make temp and data double plot
def pltd(x1,y1,x2,y2,xlabel='',ylabel1='',ylabel2='',title='',\
         ystyle1='b-',ystyle2='r-',ylim1='',ylim2='',label1='',label2=''):
    fig = figure()
    ax1 = fig.add_subplot(111)
    ax1.plot(x1,y1,ystyle1,label=label1)
    ax1.set_xlabel(xlabel)
    ax1.set_ylabel(ylabel1,color='b')
    for t1 in ax1.get_yticklabels():
        t1.set_color('b')
    if ylim1!='':
        ax1.set_ylim(ylim1)
    ax2 = ax1.twinx()
    ax2.plot(x2,y2,ystyle2,label=label2)
    ax2.set_ylabel(ylabel2,color='r')
    for t1 in ax2.get_yticklabels():
        t1.set_color('r')
    if ylim2!='':
        ax2.set_ylim(ylim2)
    ax1.set_title(title)

    return ax1,ax2

#Make temp and data double plot
def pltd2(x1,y1,fn,xlabel='',ylabel1='',ylabel2='',title='',\
         ystyle1='b-',ystyle2='r-',ylim1='',ylim2='',label1='',label2=''):
    fig = figure()
    ax1 = fig.add_subplot(111)
    ax1.plot(x1,y1,ystyle1,label=label1)
    ax1.set_xlabel(xlabel)
    ax1.set_ylabel(ylabel1)
    #Get x tick positions
    b = ax1.get_xticks()
    newb = fn(b)
    #newb = newb.astype('str')
    for i in range(size(newb)):
        newb[i] = '%.2f' % newb[i]
    #Create second axis
    ax2 = ax1.twiny()
    ax2.set_xticks(b)
    ax2.set_xticklabels(newb.astype('str'))
    
    return ax1,ax2

#Print a number in scientific notation
def scinot(x):
    print ('%e' % x)

#Find index where vector is closest to a value
def mindiff(x,y):
    diff = abs(x-y)
    diff2 = diff[invert(isnan(diff))]
    return where(diff==min(diff2))[0]

#Make a contour plot of an image
def mycontour(img,x=0\
               ,y=0,log=False,fmt=None,nlev=None,maxl=None,minl=None):
    if nlev is None:
        nlev = 100.
    if size(x) == 1:
        x = arange(shape(img)[1])
        y = arange(shape(img)[0])
    if log==False:
        if maxl==None:
            levels = linspace(nanmin(img),nanmax(img),nlev)
        else:
            levels = linspace(minl,maxl,nlev)
        c = contourf(x,y,img,levels=levels)
    else:
        mn = np.log10(nanmin(img))
        mx = np.log10(nanmax(img))
        levels = linspace(mn,mx,nlev)
        levels = 10**levels
        c = contourf(x,y,img,levels=levels,norm=colors.LogNorm())
    if fmt==None:
        fmt='%.'+str(np.int(-np.floor(log10(levels[1]-levels[0])/nlev)))+'f'
    colorbar(format=fmt)
    return c

#Convert from bin edges to bin centers
def edgestocent(x):
    c = []
    for i in range(size(x)-1):
        c.append((x[i+1]+x[i])/2.)
    return array(c)

#Return center of bins in histogram function
def myhist(arr,bins=10,density=None,range=None):
    val = histogram(arr,bins=bins,density=density,range=range)
    b = np.zeros(size(val[1])-1)
    for i in np.arange(np.size(b)):
        b[i] = val[1][i]+(val[1][1]-val[1][0])/2.
    return (val[0],b)

#Return mean of array with nans
def nanmean(arr):
    nanmask = isnan(arr)
    return arr[invert(nanmask)].mean()

#Load from pickle file
def pload(filename):
    f = open(filename,'r')
    b = pickle.load(f)
    f.close()
    return b

#Save to pickle file
def psave(data,filename):
    f = open(filename,'w')
    pickle.dump(data,f)
    f.close()



def namelist_from_string(l,separator="="):
    """extract namelist parameters from a list of strings (e.g. file lines). """
    l=[ll.strip() for ll in l if ll.count(separator) == 1 ]  #consider only valid line (one and only one = sign in the line).
    dict={}
    for ll in l:
        s=ll.split(separator)
        dict[s[0].strip()]=s[1].strip()
    return dict    


def read_pars_from_namelist(filename,separator="="):
    '''Read a set of parameters from a (fortran-like) namelist in file FILENAME.
    Return a dictionary. All values are read as strings.
    A file-object like io.StringIO(filetext) can be used in place of a real file.'''
    
    if hasattr(filename,'readlines'):
        l=filename.readlines()
    else:
        l=open(filename,'r').readlines()

    return namelist_from_string(l,separator=separator)

if __name__=="__main__":
    file=r'E:\work\WTDf\geoSettings.txt'
    d=read_pars_from_namelist(file)
    print(d)


import numpy as np
from dataIO.span import span
import pdb
import logging
import warnings

warnings.simplefilter("once")

def stats (data=None,units=None,string=False,fmt=None,vars=None):
    """ Return selected statistics on data as numerical array or list of strings (one for each stats).  
    
    vars is a list of indices that select the variables to be included, wrt a list (if called without `data` returns a format representation of the variables in the list):
    0 - mean
    1 - stddev
    2 - rms
    3 - PV
    4 - min
    5 - max
    6 - number of elements
    
    N.B.: (1) is intended as the rms of the deviation from the mean, while (2) is the root mean square of the signal as value (wrt to zero).
    Note that span doesn't exclude nan data, put flag to tune this option.
    
    string if set to True return stats as strings. In this case a string `units` can be used to add a postfix to statistics. A finer control can be obtained by passing in `fmt` a list of format strings for each var.
    e.g.: the default is obtained with:
    
            fmt = ['mean: %.5g'+units,
                   'StdDev: %.5g'+units,
                   'rms: %.5g'+units,
                   'PV: %.5g'+units,
                   'min: %.5g'+units,
                   'max: %.5g'+units,
                   'n:  %i']
    
    2021/06/30 added rms (different from standard dev, which is centered about mean).
    """

    if vars is None: vars = [0,1,2,3,4,5,6]
    
    if data is not None:
        # `st` stats to return are filtered based on vars
        data = np.array(data)
        st = [np.nanmean(data),np.nanstd(data),
              np.sqrt(np.nansum(data**2)/np.size(np.where(~np.isnan(data))[0])),
              span(data,size=True),*span(data),np.size(data)]
        if len(vars) > 0:
            st = [s for i,s in enumerate(st) if i in vars]
        else:
            st = []
    
    # convert to string
    #pdb.set_trace()
    if string or data is None:
        # `units` must be a single string and it is applied according to default fmt.   
        # `units` defaults to empty string, for different units must be incorporated in `fmt`. 
        if units is None or len(units)==0:   # if not empty adjust spacing  
            units = ""
        else:
            units = " " + units 
        if fmt is None or data is None or not(fmt):
            fmt = ['mean: %.5g'+units,
                'StdDev: %.5g'+units,
                'rms: %.5g'+units,
                'PV: %.5g'+units,
                'min: %.5g'+units,
                'max: %.5g'+units,
                'n:  %i']
            
            # filter `fmt` for indices in vars
            fmt = [f for i,f in enumerate(fmt) if i in vars]
            
            if data is None:
                warnings.warn ("`stats` called without data, print and returns default vars.")
                return fmt    
        elif np.size(fmt) == 1:  #??
            #if single string replicates to all variables
            fmt = np.repeat(fmt,len(st))
            if units:
                fmt = [f+units for f in fmt]
            
        #otherwise `fmt` and  is assumend already of same size as `st`
        
        st = [f%s for f,s in zip(fmt,st)]
        # 2021/06/30 here st and fmt are of same length. 
        # before today it was expecting complete format
        # st = [s for i,s in enumerate(st) if i in vars]  # filter st
    
    return st

def split_on_indices(vector,indices):
    """Split an array on a list of lengths (indices represent the number of items in each block). 
    Return a list of splitted sub-arrays. """
    
    if np.sum(indices) != len(vector):
        Warning('len of indices for splitting is not matching vector length!')
    ind = np.cumsum(indices).tolist()
    result = [vector[i1:i2] for i1,i2 in zip([0]+ind,ind)]

    return result


def split_blocks (vector, sep = None, direction = None, split = False):

    """ Given a vector return the lenght of blocks (or blocks if split is True).
    
    A block is a monotonic sequence.
    
    The direction of monotony can be explicitly set with ``direction`` argument 
    (positive if increasing, negative if decreasing, 0 consider blocks as formed by equal elements).
    If direction is not provided, it is inferred from the first two elements.
    
    Return a list: if ``split`` is False (default), a list of the lengths of blocks is returned, otherwise a list of blocks is returned. 
    
    (e.g. [1,2,3,1,2,3] -> return [3,3])
    (e.g. [1,1,1,2,2,2,3,3,3,4,4,4], equal = True -> return [3,3,3,3]) """
    
    if direction is None:  # exclude invalid points
        x = [vector[i] for i in np.where(np.isfinite(vector))[0]]
        direction = np.sign(x[1]-x[0])
    diff = np.diff(vector)
    
    if direction == 0:
        diff = np.insert(diff,0,0)
        ind = np.where(diff != 0)[0].tolist()
    else:
        direction = np.sign(direction)
        ind = list(np.where(np.sign(diff) != np.sign(direction))[0]+1)
    ind.append (len(vector))
    
    # correct for nan or invalid data, which artificially create two artificial changes of direction.
    # keep only the first, invalid data will be included in output in second block.
    ind = [i for i in ind if i in np.where(np.isfinite(vector))[0]+1]  # +1 to include nan in following block and keep into account first element if invalid. I suspect it will give problems if last element is invalid,
    #must be kept specifically into account.
    #if not np.isfinite(vector[0]): ind.remove(1)  # 1 is legitimate only if first two elements are finite
    
    #if not cumulative or split: # cumulative can be obtained with np.cumsum()
    ind = [i1-i2 for i1,i2 in zip(ind,[0]+ind[:-1])]
    
    
    
    return split_on_indices(vector, ind) if split else ind


def test_len_blocks():

    print(split_blocks([1,2,3,1,2,3]))   #[3,3]
    print(split_blocks([1,2,3,1,2]))   #[3,2]
    print(split_blocks([1,2,3,1,2],direction = -1))   #[1,1,2,1]
    print(split_blocks([1,2,3,1,2],direction = 0))   #[1,1,1,1,1]
    print(split_blocks([1,1,1,2,2,3,3,3],direction = 0))   #[3,2,3]
    
    print(split_blocks([1,2,3,1,2,3],split = True))   #[[1,2,3],[1,2,3]]
    print(split_blocks([1,2,3,1,2],split = True))   # [[1, 2, 3], [1, 2]]
    print(split_blocks([1,2,3,1,2],direction = -1,split = True))   # [[1], [2], [3, 1], [2]]
    print(split_blocks([1,2,3,1,2],direction = 0,split = True))   # [[1], [2], [3], [1], [2]]
    print(split_blocks([1,1,1,2,2,3,3,3],direction = 0,split = True))   # [[1, 1, 1], [2, 2], [3, 3, 3]]



def test_stats(*args,**kwargs):
    # generate random distribution
    n  = 1000
    
    v = np.random.random(n)
    
    print('len(data): ',len(v))   
    print("no options: all stats as float:\n")
    print(stats(v),"\n----------\n\n")    
    print("with empty vars, return empty:", stats(v,vars=[]),"\n----------\n\n")
    print("select two stats (vars=[0,1])", stats(v,vars=[0,1]),"\n----------\n\n")
    
    print("return as string (string=True), include units:\n", stats(v,string=True,vars = [2,6], units ='km'),"\n----------\n\n")    
    print("return as string (single string), that is appended to format according to default meaning:", stats(v,string=True,vars = [2,6],fmt = ['rms: %.6g A',
    'size: %.3i pts'],units='[A]'),"\n----------\n\n")  
    print("for finer control (e.g. more that one unit), it needs to be included in fmt)", stats(v,string=True,vars = [2,6],fmt = ['rms: %.6g A',
    'N: %.3i points']),"\n----------\n\n")
    
    print("if called without data, return a format string (units can be added, fmt is irrelevant)", stats(string=True,fmt=None,units='um',vars = [2,6]),"\n----------\n\n") 
    
    print("without data and vars return full format string", stats(string=True),"\n----------\n\n") 
 
    return stats(v,*args,**kwargs)
    
def is_nested_list(l):
    """Return true if it is a nested list (at least one element is a list), False otherwise (all elements are scalar).
    Uses isinstance, so arrays are not considered lists. 
    from https://stackoverflow.com/questions/24180879/python-check-if-a-list-is-nested-or-not."""
    #breakpoint()
    try:
          next(x for x in l if isinstance(x,list))
    
    except StopIteration:
        return False
    except TypeError: #l is scalar or None
        return False
    
    return True

def test_is_nested_list():
    """
    testvalues=[None,[None],[None,[],[]],[[1,2,3]],[3],[[3]],[[],[2],[3]],[[1],[2],[3]],[[1],[2],[3]],[[1],[2,3],[3]],[[2,3],[],[3]]]
    '''
    None | False
    [None] | False
    [None, [], []] | True
    [[1, 2, 3]] | True
    [3] | False
    [[3]] | True
    [[], [2], [3]] | True
    [[1], [2], [3]] | True
    [[1], [2], [3]] | True
    [[1], [2, 3], [3]] | True
    [[2, 3], [], [3]] | True
    '''
    """
    testvalues=[None,[None],3,[None,[],[]],[[1,2,3]],[3],[[3]],
        [[],[2],[3]],[[1],[2],[3]],[[1],[2],[3]],
        [[1],[2,3],[3]],[[2,3],[],[3]]]
    for tv in testvalues:
        #print('test string output:')
        print (tv,"|",is_nested_list(tv))
        
    
if __name__ == "__main__":
    test_stats()


''' Vincenzo Cotroneo vcotroneo@cfa.harvard.edu 2016/03/07

It's a set of classes that add key commands to a matplotlib window.
The class is created and called. On call with an object and a figure as arguments,
links them and add the key commands to the matplotlib figure.

There are 

class BaseImageKey(AbstractImageKey):
    """A simple one with only help and reset (plt.close('all')).
    can be used as base."""

class BaseImageKey(AbstractImageKey):
    """A simple one with only help and reset (plt.close('all')).
    can be used as base."""

    
2018/06/06 it used to work, but dependent on Tk , not on my computer.
It was used to visualize help.
Also, some problem with arguments passed to event functions for BaseImageKey
example b.
'''

import matplotlib
tk_yes=False  #set to true if tkinter is present
try:

    #http://effbot.org/tkinterbook/tkinter-file-dialogs.htm
    import tkinter as tk
    import tkinter.filedialog as filedialog
    import tkinter.simpledialog
    from tkinter import Frame,Label,Entry,Button
    import tkinter.messagebox
    matplotlib.use('TkAgg')
    tk_yes=True
except ImportError:
    print("tkinter not laoded")
    pass
import matplotlib.pyplot as plt 


from dataIO.fn_add_subfix import fn_add_subfix
import re
import shutil
import os
import numpy as np

from dataIO.span import span 


import pdb
  

   
pe_yes=False  #set to true if piexif is present
try:
    import piexif
    pe_yes=True
except ImportError:
    pass

''' Following classes are minimal examples built for debugging, 
not needed any more after AbstractImageKey class works properly.

class minimalExample(object):
    """Try to connect to a function."""
        
    def __call__(self):
        """connect the current figure to an object adding I/O command interface based on keyboard to figure.
        object is passed to be manipulated in routine. Includes standard keys
        use H for list of commands."""
        
        def on_key_press(event):
            print "pressed"
            plt.clf()
            plt.draw()
            pass
                
        figure=plt.gcf()                               
        ax=plt.gca()
        
        print "connect"
        plt.connect('key_press_event', on_key_press)
            
        return figure  #return

class minimalExample2(object):
    """Try to connect to a function."""

  
    def __init__(self):
        def pltrnd():
            plt.plot(np.arange(100),np.random.rand(100))
            plt.draw()
        self.cmds={'c':plt.clf,'r':pltrnd}
        
    def __call__(self):
        """connect the current figure to an object adding I/O command interface based on keyboard to figure.
        object is passed to be manipulated in routine. Includes standard keys
        use H for list of commands."""
            
        def on_key_press(event):
            print "pressed"+event.key
            if event.key in self.cmds.keys(): 
                self.cmds[event.key]()
                print event.key+'!'
            plt.draw()
                
        figure=plt.gcf()                               
        ax=plt.gca()
        
        print "connect"
        plt.connect('key_press_event', on_key_press)
            
        return figure  #return
'''
     
class AbstractImageKey(object):
    """It's a class that when called connects a set of key shortcuts and functions
    to a figure. 
     
    Empty class to be inherited from.
    figure is not stored in any property, the ImageKey object defines the 
        keyboard interface.
    It is initialized with three lists of keys, functions and labels.

    For each of them a key is added connected to figure as shortcut to activate the function.
    In this version, function has self as argumennt: evt are kept in this class
        and in its derived.
    Label is used to generate an entry in the help dictionary, the help function
    is not defined here.
    functions and keys can be added with method add_key,
    if called return figure (unless arguments are provided, see connect).
    
    connect is called when the instance is called as a function.
    """
    
    def __init__(self,keys=None,functions=None,labels=None):
        self.cmds={}
        self.helpstrings={}
        # Deactivate the default keymap
        fig=plt.gcf()
        fig.canvas.mpl_disconnect(fig.canvas.manager.key_press_handler_id)
        if keys and functions and labels:
            for k,f,l in zip(keys,functions,labels):
                self.add_key(k,f,l)
        
    def add_key(self,key,function,label=None):
        if label is None:
            label=key+': '+function +'(?)'
        self.cmds[key]=function
        self.helpstrings[key]=label  
        
    def connect(self,obj=None,figure=None):
        """connect the current figure to an object adding I/O command interface based on keyboard to figure.
        object is passed to be manipulated in routine. (?)
        figure can be a plt.figure, otherwise current figure is used.
        return obj if not None, otherwise figure, I don't know why.
        It is called when self is called.
        
        """
        
        def on_key_press(evt):
            """call the corresponding function in self.cmds try both
            passing evt and witout, not sure it always work.
            it is connected to matplotlilb at the end"""
            #print ("pressed"+evt.key)
            #print (self.cmds)
            #print (evt.key in self.cmds.keys())
            if evt.key in self.cmds: 
                #print (self.cmds[evt.key])
                print ('call')
                try:
                    self.cmds[evt.key]()  
                except TypeError: 
                    #print('cane')
                    self.cmds[evt.key](self)
                    #problem is here the cmd can accept
                        #evt if needed to extract info, but can also be without
                        #parametner, how do I solve? use this limited workaround
                #except Exception as err:
                #    raise err
                plt.draw()                
        
        def handle_close(evt):
            """Useful maybe to save log? At the moment is saved in quit."""
            print('Closed Figure!')
                
        if figure is None: 
            figure=plt.gcf()
        else:
            plt.figure(figure.number)
        
        #ax=plt.gca()
                
        plt.connect('key_press_event', on_key_press)
        plt.connect('close_event', handle_close) 
            
        if obj is None:
            obj=figure
            
        return obj  #return

    def help_str(self):
        #import pdb
        #pdb.set_trace()
        #helpstrings=["-KEY COMMANDS-"].extend(['%s: %s'%(key,self.helpstrings[key]) for key in list(self.cmds.keys())])
        helpstrings=["-KEY COMMANDS-"]+['%s: %s'%(key,self.helpstrings[key]) for key in self.cmds]
        
        helptext= "\n".join(helpstrings)
        return helptext 
        
    def __call__(self,obj=None,figure=None):
        self.connect(obj,figure)
        
        
class BaseImageKey(AbstractImageKey):
    """A simple one with only help and reset (plt.close('all')).
    can be used as base.
    Adds tk interface to Abstract."""

    def __init__(self,*args,**kwargs):
    
        #def call_reset(evt):
        #    plt.close('all')
        
        def call_help(evt=None):
            helptext=self.help_str()  #calls parent self generated string
            try:            
                root = tk.Tk()
                #root.withdraw()
                w = tk.Label(root, text=helptext)
                w.pack()
                root.mainloop()            
            except Exception as err:
                raise err                
                #plt.connect(lambda x: print(help_str(x)))

        super(BaseImageKey, self).__init__(*args,**kwargs)
        #self.add_key('r',call_reset,'reset all plots')
        self.add_key('h',call_help,'Show this help')
        try:
            tk.Tk
        except NameError:
            print('backend_error')
        
    def __call__(self,object=None,figure=None):
        """
        This is very confused.
        connect the current figure to an object adding I/O command interface based on keyboard to figure.
        object is passed to be manipulated in routine. Includes standard keys
        use H for list of commands.
        If object is None is set to current axis (?!) call connect of abstract class
        with object and figure, then return obj."""
             
        def endroot(root):  #also root.destroy() ?
            root.quit()
            
        def handle_close(evt):
            """Useful maybe to save log? At the moment is saved in quit."""
            print('Closed Figure!')
                
        if figure is None: 
            figure=plt.gcf()
        else:
            plt.figure(figure.number)
                               
        ax=plt.gca()            
        if object is None:
            object==ax
            
        self.connect(object,figure)
        return object  #return the same object. It was figure,ax in SampleImageList, where nearly all code used plt.gca()
        
    
"""
http://stackoverflow.com/questions/7969352/matplotlib-interactively-select-points-or-locations
from __future__ import print_function
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from matplotlib.patches import Rectangle
from matplotlib.text import Text
from matplotlib.image import AxesImage
import numpy as np
from numpy.random import rand

if 1:
    fig, ax = plt.subplots()
    ax.set_title('click on points', picker=True)
    ax.set_ylabel('ylabel', picker=True, bbox=dict(facecolor='red'))
    line, = ax.plot(rand(100), 'o', picker=5)

    def onpick1(event):
        if isinstance(event.artist, Line2D):
            thisline = event.artist
            xdata = thisline.get_xdata()
            ydata = thisline.get_ydata()
            ind = event.ind
            print 'X='+str(np.take(xdata, ind)[0]) # Print X point
            print 'Y='+str(np.take(ydata, ind)[0]) # Print Y point

    fig.canvas.mpl_connect('pick_event', onpick1)
"""

if __name__=="__main__":

    def pltrnd():
        """a test function that plots a random array on current axis."""
        plt.plot(np.arange(100),np.random.rand(100))
        plt.draw()
        print('pltrnd!')
        
    def display_help(self):
        """builds a help string from a list of """
        #helpstrings=["-KEY COMMANDS-"]+['%s: %s'%(key,self.helpstrings[key]) for key in list(self.cmds.keys())]
        #helptext= "\n".join(helpstrings)
        helptext=self.help_str()
        
        print (helptext)
        root = tk.Tk()
        root.lift()
        root.focus_force()
        #root.withdraw()
        w = tk.Label(root, text=helptext)
        w.pack()
        root.mainloop()
        
    switch='010'
        
    plt.close('all')    
    
    if switch[0] != '0':
        # example on using a AbstractImageKey (completely blank class)
        # and add some functions
        pltrnd() #make a plot 
        # create the key commands and 
        a=AbstractImageKey(['c','r'],[plt.clf,pltrnd],['clear figure','plot random image'])
        #a()    
        a.add_key('h',display_help,'display this help')  ##here I want to add a function that
        a()
        plt.title('AbstractImageKey with added c, r, h')
        #plt.ion()
        plt.show()

    if switch[1] != '0':    
        #same with BaseImageKey, here uses standard keys
        plt.figure()
        pltrnd() #make a plot 
        b=BaseImageKey()
        b.add_key('r',pltrnd,'plot random line') 
        b()
        plt.title('BaseImageKey with standard h, added r')
        plt.show()
        
    if switch[2] != '0':
        plt.figure()
        pltrnd() #make a plot 
        c=BaseImageKey()
        c()
        plt.title('BaseImageKey with standard h, r')
        plt.show()        



#from pprint import pprint
import numpy as np
import os

import warnings
warnings.filterwarnings("error", category=np.VisibleDeprecationWarning) 

def read_blocks(filename,delimiter = '\n\n', comment = None,*args,**kwargs):
    """Read a file containing blocks of numerical data separated by a white line.
    
    Return a list, each element of which is a list of the lines in the block.  
    2022/06/29 converted result from dictionary to list.    
    """
    
    f = open(filename, 'r')
    content = f.read()
    f.close()
    pList=[]
    for block in content.split(delimiter):
        if (block.strip() != ''):
            """build a dic with a set of points, first line is used as key"""
            lines=block.split('\n')
            goodlines = [dd for dd in lines if (comment is None) or not(dd.startswith(comment))] 
            goodlines = [gl for gl in goodlines if len(gl.strip())!=0] #il caso di commento e' gia' escluso
            try:
                l = np.array([dd.split() for dd in goodlines],*args,**kwargs)
                l = np.array([dd.split() for dd in goodlines],*args,**kwargs)
            except np.VisibleDeprecationWarning:
                print ('Triying to convert splitted strings with unknown dtype results'+
                       ' in non corresponding lengths. Return array of strings.')
                l = goodlines
            except ValueError:
                print('Data type not corresponding to user-provided dtype. Return array of strings.')
                l = goodlines
                
            pList.append(np.array(l))

    return pList
    
def test_read_blocks(filename=None):

    def print_results(a, title):
        '''pass result and descriptive title for a test.'''
        
        print('\n======',title)
        print('%i blocks read'%len(a))
        print('block shapes: ',[aa.shape for aa in a])
        print('block contents: ')
        for i,aa in enumerate(a):
            print('%i)\n'%(i+1),'[%s\n..\n..\n%s]'%(','.join(map(str,aa[:3])),','.join(map(str,aa[-2:]))))
        return a
    
    filename = r'C:\Users\kovor\Documents\python\pyXTel\source\pyProfile\test\input_data\data_blocks_spaced.dat'
    
    print('read "%s", contains three white-line-separated blocks with initial 3-line # comment followed by 94 couples of data. First block has 6-line comment.'%os.path.join(os.path.basename(os.path.dirname(filename)),os.path.basename(filename)))
    
    a = print_results(read_blocks(filename, comment = '#'),"comment = '#', ignore commented lines, lines are splitted. Return 2xN string array.") #array of strings 2 x N
    
    a = print_results(read_blocks(filename, comment = '#', dtype = float),"comment = '#' with float dtype, Return 2xN float array.") #array of floats 2 x N
    
    a = print_results(read_blocks(filename),"no comment or dtype, lines are splitted but they are inconsistent length. Return string vector with lines.")  # da warning in quanto lo split funziona creando liste. Se queste sono di lunghezze diversa,
                         # la conversione in array restituiscce un array monodimensionale (vettore) di oggetti (lista)

    a = print_results(read_blocks(filename, dtype = float),"dtype, but no comment character, comment lines break consistency. Return string vector with lines.") #lista di array
    
    
    
    
def head(fn,N=10):
    """return first n lines of file `fn`, without reading the other lines.
    
    from https://stackoverflow.com/questions/1767513/read-first-n-lines-of-a-file-in-python"""
    with open(fn) as myfile:
        return [next(myfile) for x in range(N)]
    

def program_path0():
    """first version. Both are from 
    https://stackoverflow.com/questions/51487645/get-path-of-script-containing-the-calling-function
    this one is not working on case 2 of test.
    See also:
    https://note.nkmk.me/en/python-script-file-path/#:~:text=In%20Python%2C%20you%20can%20get,python%20(or%20python3%20)%20command.
    """
    import inspect
    stack = inspect.stack()
    calling_context = next(context for context in stack if context.filename != __file__)
    return calling_context.filename
    #return stack


def program_path():
    """return the path
    
    to the calling file. """
    import sys

    namespace = sys._getframe(1).f_globals  # caller's globals
    #pprint(namespace)
    return namespace['__file__']


def test_program_path():
    """cases:
        1. called from input (VScode python file): 
            <ipython-input-11-0fddf82ac4ab>     
            
        if 2. called from ipython shell running this file:
            %run .../pyXTel/dataIO/files.py
            
        or 3. importing from this function from here with:
            from dataIO.files import test_program_path
            test_program_path()
        
        gives correct result:            
            caller's path: ...\pyXTel\dataIO\files.py
            

       if 4. called from a  .py file containing a copy of 
            this function.
            
            %run .../dataIO/test/test_program_path.py
            
            caller's path: ...\dataIO\test\test_program_path.py
            
        N.B. program_path0 gives incorrect results on case 2. only.
    """ 
    p = program_path0()  
    print("caller's path: ", p )
    return  p
    
    
        
if __name__ == "__main__":
    a = test_program_path()


# 2020/10/25 moved here from dataIO, not sure why it was there.
from dataIO.span import span

def external_roi_rect(points):
    """
    finds the smallest (straight to axis) rectangle containing all points.
    
    Points are passed as N x 2 array, returns two couples (x0,x1) (y0,y1)
    """
    return span(points,axis=0)


def in_hull(p, hull):
    """
    Test if points in `p` are in `hull`

    `p` should be a `NxK` coordinates of `N` points in `K` dimensions
    `hull` is either a scipy.spatial.Delaunay object or the `MxK` array of the 
    coordinates of `M` points in `K`dimensions for which Delaunay triangulation
    will be computed
    """
    from scipy.spatial import Delaunay
    if not isinstance(hull,Delaunay):
        hull = Delaunay(hull)

    return hull.find_simplex(p)>=0
    



''' Vincenzo Cotroneo vcotroneo@cfa.harvard.edu 2016/03/07

It's a set of classes that add key commands to a matplotlib window.
The class is created and called. On call with an object and a figure as arguments,
links them and add the key commands to the matplotlib figure.

There are 

class BaseImageKey(AbstractImageKey):
    """A simple one with only help and reset (plt.close('all')).
    can be used as base."""

class BaseImageKey(AbstractImageKey):
    """A simple one with only help and reset (plt.close('all')).
    can be used as base."""
    
2018/06/06 it was used to work, but dependent on Tk , not on my computer.
It was used to visualize help.
Also, some problem with arguments passed to event functions for BaseImageKey
example b.

'''


from dataIO.fn_add_subfix import fn_add_subfix
import re
import shutil
import os
import numpy as np
import matplotlib.pyplot as plt
from dataIO.span import span 
import matplotlib
import pdb


tk_yes=False  #set to true if tkinter is present
try:
    #http://effbot.org/tkinterbook/tkinter-file-dialogs.htm
    import tkinter as tk
    import tkinter.filedialog as filedialog
    import tkinter.simpledialog
    from tkinter import Frame,Label,Entry,Button
    import tkinter.messagebox
    matplotlib.use('TkAgg')
    tk_yes=True
except ImportError:
    print("tkinter not laoded")
    pass

pe_yes=False  #set to true if piexif is present
try:
    import piexif
    pe_yes=True
except ImportError:
    pass

''' Following classes are minimal examples built for debugging, 
not needed any more after AbstractImageKey class works properly.

class minimalExample(object):
    """Try to connect to a function."""
        
    def __call__(self):
        """connect the current figure to an object adding I/O command interface based on keyboard to figure.
        object is passed to be manipulated in routine. Includes standard keys
        use H for list of commands."""
        
        def on_key_press(event):
            print "pressed"
            plt.clf()
            plt.draw()
            pass
                
        figure=plt.gcf()                               
        ax=plt.gca()
        
        print "connect"
        plt.connect('key_press_event', on_key_press)
            
        return figure  #return

class minimalExample2(object):
    """Try to connect to a function."""

  
    def __init__(self):
        def pltrnd():
            plt.plot(np.arange(100),np.random.rand(100))
            plt.draw()
        self.cmds={'c':plt.clf,'r':pltrnd}
        
    def __call__(self):
        """connect the current figure to an object adding I/O command interface based on keyboard to figure.
        object is passed to be manipulated in routine. Includes standard keys
        use H for list of commands."""
            
        def on_key_press(event):
            print "pressed"+event.key
            if event.key in self.cmds.keys(): 
                self.cmds[event.key]()
                print event.key+'!'
            plt.draw()
                
        figure=plt.gcf()                               
        ax=plt.gca()
        
        print "connect"
        plt.connect('key_press_event', on_key_press)
            
        return figure  #return
'''
     
class AbstractImageKey(object):
    """It's a function that returns references to figure and axis after adding a key shortcut to the axis.
    Empty class to be inherited from."""
    
    def __init__(self,keys=None,functions=None,labels=None):
        self.cmds={}
        self.helpstrings={}
        # Deactivate the default keymap
        fig=plt.gcf()
        fig.canvas.mpl_disconnect(fig.canvas.manager.key_press_handler_id)
        if keys and functions and labels:
            for k,f,l in zip(keys,functions,labels):
                self.add_key(k,f,l)
        
    def add_key(self,key,function,label=None):
        if label is None:
            label=key+': '+function +'(?)'
        self.cmds[key]=function
        self.helpstrings[key]=label  
        
    def connect(self,object=None,figure=None):
        """connect the current figure to an object adding I/O command interface based on keyboard to figure.
        object is passed to be manipulated in routine. Includes standard keys
        use H for list of commands."""
        
        def on_key_press(event):
            #print "pressed"+event.key
            #print self.cmds
            #print event.key in self.cmds.keys()
            if event.key in list(self.cmds.keys()): 
                #print self.cmds[event.key]
                self.cmds[event.key]()    
                plt.draw()                
        
        def handle_close(evt):
            """Useful maybe to save log? At the moment is saved in quit."""
            print('Closed Figure!')
                
        if figure is None: 
            figure=plt.gcf()
        else:
            plt.figure(figure.number)
        
        #ax=plt.gca()
                
        plt.connect('key_press_event', on_key_press)
        plt.connect('close_event', handle_close) 
            
        if object is None:
            object=figure
            
        return object  #return
        
    def __call__(self,object=None,figure=None):
        self.connect(object,figure)
        
        
class BaseImageKey(AbstractImageKey):
    """A simple one with only help and reset (plt.close('all')).
    can be used as base."""

    def __init__(self,*args,**kwargs):
    
        def call_reset(evt):
            plt.close('all')
        
        def call_help(evt):
            helptext= "\n".join(helpstrings)
            root = tk.Tk()
            #root.withdraw()
            w = tk.Label(root, text=helptext)
            w.pack()
            root.mainloop()            

        super(BaseImageKey, self).__init__(*args,**kwargs)
        self.add_key('r',call_reset,'reset plots')
        self.add_key('h',call_help,'Show this help')
        
    def __call__(self,object=None,figure=None):
        """connect the current figure to an object adding I/O command interface based on keyboard to figure.
        object is passed to be manipulated in routine. Includes standard keys
        use H for list of commands."""
             
        def endroot(root):
            root.quit()
        def handle_close(evt):
            """Useful maybe to save log? At the moment is saved in quit."""
            print('Closed Figure!')
                
        if figure is None: 
            figure=plt.gcf()
        else:
            plt.figure(figure.number)
                               
        ax=plt.gca()            
        if object is None:
            object==ax
            
        self.connect(object,figure)
        return object  #return the same object. It was figure,ax in SampleImageList, where nearly all code used plt.gca()
        
    
"""
http://stackoverflow.com/questions/7969352/matplotlib-interactively-select-points-or-locations
from __future__ import print_function
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from matplotlib.patches import Rectangle
from matplotlib.text import Text
from matplotlib.image import AxesImage
import numpy as np
from numpy.random import rand

if 1:
    fig, ax = plt.subplots()
    ax.set_title('click on points', picker=True)
    ax.set_ylabel('ylabel', picker=True, bbox=dict(facecolor='red'))
    line, = ax.plot(rand(100), 'o', picker=5)

    def onpick1(event):
        if isinstance(event.artist, Line2D):
            thisline = event.artist
            xdata = thisline.get_xdata()
            ydata = thisline.get_ydata()
            ind = event.ind
            print 'X='+str(np.take(xdata, ind)[0]) # Print X point
            print 'Y='+str(np.take(ydata, ind)[0]) # Print Y point

    fig.canvas.mpl_connect('pick_event', onpick1)
"""

if __name__=="__main__":

    def pltrnd():
        """a test function that plots a random array on current axis."""
        plt.plot(np.arange(100),np.random.rand(100))
        plt.draw()
        print('pltrnd!')
        
    def display_help(self):
        """builds a help string from a list of """
        helpstrings=["-KEY COMMANDS-"].extend(['%s: %s'%(key,self.helpstrings[key]) for key in list(self.cmds.keys())])
        helptext= "\n".join(helpstrings)
        root = tk.Tk()
        #root.withdraw()
        w = tk.Label(root, text=helptext)
        w.pack()
        root.mainloop()
        
    # example on using a AbstractImageKey (completely blank class)
    pltrnd() #make a plot 
    # create the key commands and 
    a=AbstractImageKey(['c','r'],[plt.clf,pltrnd],['clear figure','plot random image'])
    #a()    
    a.add_key('h',display_help,'display this help')  ##here I want to add a function that
    a()
    #plt.ion()
    plt.show(block=True)
    
    #same with BaseImageKey
    pltrnd() #make a plot 
    b=BaseImageKey()
    b()
    plt.show(block=True)



def rebin(a, *args):
    '''rebin ndarray data into a smaller ndarray of the same rank whose dimensions
    are factors of the original dimensions. eg. An array with 6 columns and 4 rows
    can be reduced to have 6,3,2 or 1 columns and 4,2 or 1 rows.
    example usages:
    >>> a=rand(6,4); b=rebin(a,3,2)
    >>> a=rand(6); b=rebin(a,2)
    '''
	#http://scipy-cookbook.readthedocs.org/items/Rebinning.html
    shape = a.shape
    lenShape = len(shape)
    factor = asarray(shape)/asarray(args)
    evList = ['a.reshape('] + \
             ['args[%d],factor[%d],'%(i,i) for i in range(lenShape)] + \
             [')'] + ['.sum(%d)'%(i+1) for i in range(lenShape)] + \
             ['/factor[%d]'%i for i in range(lenShape)]
    print(''.join(evList))
    return eval(''.join(evList))


import numpy as np
import matplotlib.pyplot as plt
import os
import pdb

def fread(fid, nelements, dtype):
    if dtype is str:
        dt = np.uint8  # WARNING: assuming 8-bit ASCII for np.str!
    else:
        dt = dtype
    #pdb.set_trace()
    data_array = np.fromfile(fid, dt, nelements)
    data_array.shape = nelements #(nelements, 1)
     
    return data_array

"""
%read char array until first '\0'
function s=freadChar(fileID,bytes)
s=char(fread(fileID,bytes,'char')');
tmp= strfind((s),(char(0)));
if ~isempty(tmp)
    s=s(1:tmp(1));
end
%fix encoding
s=strrep(s,'','');
end
"""     

def freadChar(fileID,n):
    a= fread(fileID,n,np.uint8)
    s= bytes(a) 
    return "".join(map(chr,s)).strip('\x00')  #strip terminal nulls  
                    #this works where s.decode().strip('\x00') fails 
                    # e.g if first character is \xb5 (greek mu for micro)
    
class Res():
    pass
    
def readsur(filepath,raw=False):
    """
    reads surface data from Surf file format 2010. returns an object with all properties from the sur file.
    Data and axes are contained respectively in res.points, res.xAxis and res.yAxis.
    
    filepath is the full path to sur file.
    `raw` flag determines if points property contains internal representation rather than rescaled and 
    reshaped values (this uses half memory space wrt original matlab function).   
    
    returns an object with all properties from the sur file.
        res.points contains reshaped and scaled or as internally represented in the sur file 
            if `raw` flag is set True.
        res.x/yAxis contains the corresponding axis.
        
    ported to python by Vincenzo Cotroneo 2018/04/18 from matlab routine by Eike Foremny, v1.0 23.02.2017
        
    """
    fileID = open(filepath, 'rb')
    res=Res()

    res.signature           =   freadChar(fileID,12);               #1
    res.format              =   fread(fileID,1,'int16');            #2
    res.objNum              =   fread(fileID,1,'int16');            #3
    res.version             =   fread(fileID,1,'int16');            #4
    res.objType             =   fread(fileID,1,'int16');            #5
    res.objName             =   freadChar(fileID,30);               #6
    res.operatorName        =   freadChar(fileID,30);               #7

    res.materialCode        =   fread(fileID,1,'int16');            #8
    res.acquisitionType     =   fread(fileID,1,'int16');            #9
    res.rangeType           =   fread(fileID,1,'int16');            #10
    res.specialPoints       =   fread(fileID,1,'int16');            #11
    res.absoluteHeights     =   fread(fileID,1,'int16');            #12
    res.gaugeResolution     =   fread(fileID,1,'single'); #fread(fileID,1,'float');            #13

    freadChar(fileID,4);        #reserved space. for unknown reason this works in matlab, not in python  #14

    #data format of pints
    res.sizeOfPoints        =   fread(fileID,1,'int16');            #15

    res.zMin                =   fread(fileID,1,'int32');            #16
    res.zMax                =   fread(fileID,1,'int32');            #17

    #Number of points per axis
    res.xPoints             =   int(fread(fileID,1,'int32'))            #18
    res.yPoints             =   int(fread(fileID,1,'int32'))            #19

    res.totalNumberOfPoints =   fread(fileID,1,'int32');            #20
    
    #Distance between points
    res.xSpacing            =   fread(fileID,1,'single') #'float'); #21
    res.ySpacing            =   fread(fileID,1,'single') #'float'); #22
    res.zSpacing            =   fread(fileID,1,'single') #'float'); #23

    #Name of axis
    res.xName               =   freadChar(fileID,16);               #24
    res.yName               =   freadChar(fileID,16);               #25
    res.zName               =   freadChar(fileID,16);               #26

    #Unit of distance between points
    res.xStepUnit           =   freadChar(fileID,16);               #27
    res.yStepUnit           =   freadChar(fileID,16);               #28
    res.zStepUnit           =   freadChar(fileID,16);               #29

    #Unit of axis
    res.xLengthUnit         =   freadChar(fileID,16);               #30
    res.yLengthUnit         =   freadChar(fileID,16);               #31
    res.zLengthUnit         =   freadChar(fileID,16);               #32

    #Skaling of distance between points
    res.xUnitRatio          =   fread(fileID,1,'single') #'float'); #33
    res.yUnitRatio          =   fread(fileID,1,'single') #'float'); #34
    res.zUnitRatio          =   fread(fileID,1,'single') #'float'); #35

    res.imprint             =   fread(fileID,1,'int16') #'float'); #36
    res.inverted            =   fread(fileID,1,'int16') #'float'); #37
    res.levelled            =   fread(fileID,1,'int16') #'float'); #38
    freadChar(fileID,12);       #obsolete                           #39

    #timestamp
    res.startSeconds        =   fread(fileID,1,'int16');            #40
    res.startMinutes        =   fread(fileID,1,'int16');            #41
    res.startHours          =   fread(fileID,1,'int16');            #42
    res.startDays           =   fread(fileID,1,'int16');            #43
    res.startMonths         =   fread(fileID,1,'int16');            #44
    res.startYears          =   fread(fileID,1,'int16');            #45
    res.startWeekDay        =   fread(fileID,1,'int16');            #46
    res.measurementDuration =   fread(fileID,1,'single') #'float'); #47

    freadChar(fileID,10);       #obsolete                           #48

    #Size of comment field
    res.commentSize         =   fread(fileID,1,'int16');            #49
    res.privateSize         =   fread(fileID,1,'int16');            #50

    res.clientZone          =   freadChar(fileID,128);              #51

    #Axis offset
    res.xOffset             =   fread(fileID,1,'single') #'float'); #52
    res.yOffset             =   fread(fileID,1,'single') #'float'); #53
    res.zOffset             =   fread(fileID,1,'single') #'float'); #54

    #temperature scale
    res.tSpacing            =   fread(fileID,1,'single') #'float'); #55
    res.tOffset             =   fread(fileID,1,'single') #'float'); #56
    res.tStepUnit           =   freadChar(fileID,13);               #57
    res.tAxisName           =   freadChar(fileID,13);               #58

    res.comment             =   freadChar(fileID,int(res.commentSize));  #59
    res.private             =   freadChar(fileID,int(res.privateSize));  #60

    #read datapoints
    if res.sizeOfPoints == 16:                                       #61
        res.points          =   fread(fileID,-1,'int16')
    elif res.sizeOfPoints == 32:
        res.points          =   fread(fileID,-1,'int32')        
    else:
        raise ValueError("data lack property sizeOfPoints (or file is in the wrong format)")

    if not(raw):
        #reshape datapoints into 2D-Matrix
        res.points       =   np.reshape(res.points,[res.yPoints,res.xPoints])   #switched x-y 20180503
                    #it was giving interlaced output with profilometer data (never noticed on square CCI images).

    #Scale datapoints without Offset
    res.points       =   (res.points-res.zMin) * res.zSpacing/ res.zUnitRatio;

    #Generate Axis without Offset;
    res.xAxis               =   np.linspace(0,res.xSpacing * res.xPoints/ res.xUnitRatio, res.xPoints);
    res.yAxis               =   np.linspace(0,res.ySpacing * res.yPoints/ res.yUnitRatio, res.yPoints);

    fileID.close();
    return res
    
if __name__=="__main__":
    
    from pySurf.data2D import plot_data
    from pySurf.readers.test_readers import testfolder
    df=os.path.join(testfolder,r'input_data\profilometer\04_test_directions\05_xysurf_pp_Height.sur')
    res=readsur(df)
    plt.figure()
    plot_data(res.points,res.xAxis,res.yAxis)
    plt.show()
    
    
    


"""experiments how it is possible to advance from a plotting function with 
interactive control as discussed in several sources on the internet."""


from __future__ import print_function
#import matplotlib
#matplotlib.use('GTK3Cairo')
#matplotlib.rcParams['toolbar'] = 'toolmanager'
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.backend_tools import ToolBase, ToolToggleBase

class ViewMarkers(ToolToggleBase):
    '''Hide lines with a given gid'''
    default_keymap = 'M'
    description = 'Hide or Show Markers'

    def __init__(self, *args, **kwargs):
        self.gid = kwargs.pop('gid')
        ToolToggleBase.__init__(self, *args, **kwargs)

    def enable(self, *args):
        self.set_lines_visibility(False)

    def disable(self, *args):
        self.set_lines_visibility(True)

    def set_lines_visibility(self, state):
        gr_lines = []
        for ax in self.figure.get_axes():
            for line in ax.markers():
                if line.get_gid() == self.gid:
                    line.set_visible(state)
        self.figure.canvas.draw()

def test_toolbar():

    #plot image and markers

    fig.canvas.manager.toolmanager.add_tool('Hide', GroupHideTool, gid='mygroup')

    # To add a custom tool to the toolbar at specific location inside
    # the navigation group
    fig.canvas.manager.toolbar.add_tool('Hide', 'navigation', 1)        


def main():
    """this works. input from keyboard, can be replaced by?
    infinite loop with check on figure property, generator.
    see e.g. https://github.com/matplotlib/matplotlib/issues/1942/"""
    plt.axis([-50,50,0,10000])
    plt.ion()
    plt.show()

    x = np.arange(-50, 51)
    for pow in range(1,5):   # plot x^1, x^2, ..., x^4
        y = [Xi**pow for Xi in x]
        plt.plot(x, y)
        plt.draw()
        plt.pause(0.01)  #this is to allow drawing. needed?
        input("Press [enter] to continue.")

def main2():

    plt.ion() # enables interactive mode
    plt.plot([1,2,3]) # result shows immediatelly (implicit draw())

    print ('continue computation')

    # at the end call show to ensure window won't close.
    plt.show()

if __name__ == '__main__':
    main2()

    



import numpy as np
import scipy
import scipy.signal
import pdb
from astropy.modeling import models,fitting

def legendre2d(d,xo=2,yo=2,xl=None,yl=None):
    """
    Fit a set of 2d Legendre polynomials to a 2D array.
    The aperture is assumed to be +-1 over each dimension.
    NaNs are automatically excluded from the analysis.
    x0 = 2 fits up to quadratic in row axis
    y0 = 3 fits up to cubic in column axis
    If xl and yl are specified, only the polynomials with
    orders xl[i],yl[i] are fitted, this allows for fitting
    only specific polynomial orders.
    """
    #Define fitting algorithm
    fit_p = fitting.LinearLSQFitter()
    #Handle xl,yl
    if xl is not None and yl is not None:
        xo,yo = max(xl),max(yl)
        p_init = models.Legendre2D(xo,yo)
        #Set all variables to fixed
        for l in range(xo+1):
            for m in range(yo+1):
                key = 'c'+str(l)+'_'+str(m)
                p_init.fixed[key] = True
        #p_init.fixed = dict.fromkeys(p_init.fixed.iterkeys(),True)
        #Allow specific orders to vary
        for i in range(len(xl)):
            key = 'c'+str(xl[i])+'_'+str(yl[i])
            p_init.fixed[key] = False
    else:
        p_init = models.Legendre2D(xo,yo)
    
    sh = np.shape(d)
    x,y = np.meshgrid(np.linspace(-1,1,sh[1]),\
                      np.linspace(-1,1,sh[0]))
    index = ~np.isnan(d)
    p = fit_p(p_init,x[index],y[index],d[index])
    
    return p(x,y),p.parameters.reshape((yo+1,xo+1))

def fitCylMisalign(d):
    """
    Fit cylindrical misalignment terms to an image
    Piston, tip, tilt, cylindrical sag, and astigmatism
    are fit
    """
    return legendre2d(d,xl=[0,0,1,2,1],yl=[0,1,0,0,1])

def fitConic(d):
    """
    Fit cylindrical misalignment terms to an image
    Piston, tip, tilt, cylindrical sag, and astigmatism
    are fit
    """
    return legendre2d(d,xl=[0,0,1,2,1,2],yl=[0,1,0,0,1,1])

def fitLegendreDistortions(d,xo=2,yo=2,xl=None,yl=None):
    """
    Fit 2D Legendre's to a distortion map as read by 4D.
    If sum of orders is odd, the coefficient needs to be negated.
    """
    #Find and format coefficient arrays
    fit = legendre2d(d,xo=xo,yo=yo,xl=xl,yl=yl)
    az,ax = np.meshgrid(list(range(xo+1)),list(range(yo+1)))
    az = az.flatten()
    ax = ax.flatten()
    coeff = fit[1].flatten()

    #Perform negation
    coeff[(az+ax)%2==1] *= -1.
    
    return [coeff,ax,az]

def sgolay2d ( z, window_size, order, derivative=None):
    """
    """
    # number of terms in the polynomial expression
    n_terms = ( order + 1 ) * ( order + 2)  / 2.0

    if  window_size % 2 == 0:
        raise ValueError('window_size must be odd')

    if window_size**2 < n_terms:
        raise ValueError('order is too high for the window size')

    half_size = window_size // 2

    # exponents of the polynomial. 
    # p(x,y) = a0 + a1*x + a2*y + a3*x^2 + a4*y^2 + a5*x*y + ... 
    # this line gives a list of two item tuple. Each tuple contains 
    # the exponents of the k-th term. First element of tuple is for x
    # second element for y.
    # Ex. exps = [(0,0), (1,0), (0,1), (2,0), (1,1), (0,2), ...]
    exps = [ (k-n, n) for k in range(order+1) for n in range(k+1) ]

    # coordinates of points
    ind = np.arange(-half_size, half_size+1, dtype=np.float64)
    dx = np.repeat( ind, window_size )
    dy = np.tile( ind, [window_size, 1]).reshape(window_size**2, )

    # build matrix of system of equation
    A = np.empty( (window_size**2, len(exps)) )
    for i, exp in enumerate( exps ):
        A[:,i] = (dx**exp[0]) * (dy**exp[1])

    # pad input array with appropriate values at the four borders
    new_shape = z.shape[0] + 2*half_size, z.shape[1] + 2*half_size
    Z = np.zeros( (new_shape) )
    # top band
    band = z[0, :]
    Z[:half_size, half_size:-half_size] =  band -  np.abs( np.flipud( z[1:half_size+1, :] ) - band )
    # bottom band
    band = z[-1, :]
    Z[-half_size:, half_size:-half_size] = band  + np.abs( np.flipud( z[-half_size-1:-1, :] )  -band )
    # left band
    band = np.tile( z[:,0].reshape(-1,1), [1,half_size])
    Z[half_size:-half_size, :half_size] = band - np.abs( np.fliplr( z[:, 1:half_size+1] ) - band )
    # right band
    band = np.tile( z[:,-1].reshape(-1,1), [1,half_size] )
    Z[half_size:-half_size, -half_size:] =  band + np.abs( np.fliplr( z[:, -half_size-1:-1] ) - band )
    # central band
    Z[half_size:-half_size, half_size:-half_size] = z

    # top left corner
    band = z[0,0]
    Z[:half_size,:half_size] = band - np.abs( np.flipud(np.fliplr(z[1:half_size+1,1:half_size+1]) ) - band )
    # bottom right corner
    band = z[-1,-1]
    Z[-half_size:,-half_size:] = band + np.abs( np.flipud(np.fliplr(z[-half_size-1:-1,-half_size-1:-1]) ) - band )

    # top right corner
    band = Z[half_size,-half_size:]
    Z[:half_size,-half_size:] = band - np.abs( np.flipud(Z[half_size+1:2*half_size+1,-half_size:]) - band )
    # bottom left corner
    band = Z[-half_size:,half_size].reshape(-1,1)
    Z[-half_size:,:half_size] = band - np.abs( np.fliplr(Z[-half_size:, half_size+1:2*half_size+1]) - band )

    # solve system and convolve
    if derivative == None:
        m = np.linalg.pinv(A)[0].reshape((window_size, -1))
        return scipy.signal.fftconvolve(Z, m, mode='valid')
    elif derivative == 'col':
        c = np.linalg.pinv(A)[1].reshape((window_size, -1))
        return scipy.signal.fftconvolve(Z, -c, mode='valid')
    elif derivative == 'row':
        r = np.linalg.pinv(A)[2].reshape((window_size, -1))
        return scipy.signal.fftconvolve(Z, -r, mode='valid')
    elif derivative == 'both':
        c = np.linalg.pinv(A)[1].reshape((window_size, -1))
        r = np.linalg.pinv(A)[2].reshape((window_size, -1))
        return scipy.signal.fftconvolve(Z, -r, mode='valid'), scipy.signal.fftconvolve(Z, -c, mode='valid')

def circle(x,y,xc,yc):
    """Fit a circle to a set of x,y coordinates
    Supply with a guess of circle center
    Returns [xc,yc],[rmsRad,rad]
    """
    fun = lambda p: circleMerit(x,y,p[0],p[1])[0]
    res = scipy.optimize.minimize(fun,np.array([xc,yc]),method='Nelder-Mead')
    return res['x'],circleMerit(x,y,res['x'][0],res['x'][1])

def circleMerit(x,y,xo,yo):
    rad = np.sqrt((x-xo)**2+(y-yo)**2)
    mrad = np.mean(rad)
    return np.sqrt(np.mean((rad-mrad)**2)),mrad



import matplotlib.pyplot as plt
import numpy as np

"""Plot markers and shapes in dependance on positions of a group of points."""

'''
def plot_poly(v,ls='',ms=''):
    """plot a polygon defined by a list of vertices `v` as coordinate couples.
    linestyle and markerstyle can be passed.
    """
    
    mh=np.array(mh)
    # duplicate first point at the end: 
    mh=np.vstack([mh,mh[0,:]]) 
    # plt.plot return a list of one element
    # if there are markers, plot them:
    if ms is not None:
        m = plt.plot(mh[:-1,0],mh[:-1,1],ms)[0]
    
    #plot lines
    return plt.plot(mh[:,0],mh[:,1],ls)[0]
'''
    
def plot_poly(v,*args,**kwargs):
    """plot a polygon defined by a list of vertices `v` as coordinate couples.
    Just a thin wrapper around `plt.plot`, it just accepts points in transpose form and adds a copy of first point at the."""
    
    v=np.array(v)
    # duplicate first point at the end: 
    v=np.vstack([v,v[0,:]]) 
    # plt.plot return a list of one element
    
    #plot lines
    return plt.plot(*v.T,*args,**kwargs)

def plot_rect(c1,c2,*args,**kwargs):
    """plot a polygon defined by a 2 element list with coordinates of opposite corners.
    linestyle and markerstyle can be passed.
    es. 
    
    2020/10/13 changed to coordinates of opposite corners from [xspan,yspan]"""
    
    mh=[[c1[0], c1[1]],[c2[0],c1[1]],[c2[0],c2[1]],[c1[0],c2[1]]]
    #mh=[[xs[0],ys[0]],[xs[1],ys[0]],[xs[1],ys[1]],[xs[0],ys[1]]]
    return plot_poly(mh,*args,**kwargs)
    
def plot_circle_ROI(c,r,color='r'):
    """plot center and circunference (color)."""
    plt.plot(c[0],c[1],'o',color=color,alpha=0.5)
    c1=plt.Circle(c,r,fill=None,color=color)
    plt.gca().add_artist(c1)
    #plt.draw()
    return c1

def circle3points(fiducials):
    """Calculate circle passing by three points.
    Return center and radius."""
    
    fid=np.array(fiducials)
    
    #plot circle on three points on wafer edge and measurement point
    x,y,z=(fid*[1,1j]).sum(axis=1)
    w = z-x
    w /= y-x
    c = (x-y)*(w-abs(w)**2)/2j/w.imag-x
    return (-c.real, -c.imag), abs(c+x)
    
def plot_circle3points(fiducials):
    """given three fiducial points, plot them and the circle passing by them.
    Return line (passing through points) and circle artists."""
    
    fid=np.array(fiducials)
    pfid=plt.plot(fid[:,0],fid[:,1],'x')
    ax=plt.gca()
    
    c,r=circle3points(fiducials)
    ax.set_aspect('equal')
    cir=plt.Circle(c,r,fill=0)
    ax.add_artist(cir)
    return pfid,cir
    
def plot_positions(fiducials,positions,labels=None,*args,**kwargs):
    """
    Obsolete: use a combinatio of plt commands, or make better separated functions to plot circles, markers, labels, etc.
    
    plot circle on three point `fiducials` and mark the three points, plot `labels` at `positions`.
    fiducials is a list of three coordinate couples.
    positions a list of n points in same format.
    label can be set to empty string to generate numeric labels (useless, can pass e.g. `[str(i).zfill(2) for i in range(N)]`).
    
    This can be replaced by a combination of markers and plot_circle3points.
    Praticamente una funzione bruttina che plotta un combinazione inutile di roba.
    Si potrebbe notevolmente migliorare.
    """
    
    pfid,cir=plot_circle3points(fiducials)
    pos=np.array(positions)    
    fs=kwargs.pop('fontsize',20)
    
    if labels is not None:    
        if labels=='':
            labels=['%3i'%ll for ll in range(pos.shape[0])]
        plab=[]
        for p,l in zip(pos,labels):
            plab.append(plt.text(p[0],p[1],l,fontsize=fs))
    
    ppos=plt.plot(pos[:,0],pos[:,1],*args,**kwargs)
    plt.draw()
    return pfid,cir,ppos,plab

if __name__=="__main__":
    fiducials=((77.312, 87.499), (77.899,102.718), (29.208,103.132))
    positions=[[58.158,97.466],
    [59.652,104.989],
    [63.947,95.724],
    [68.501,97.656],
    [69.928,88.653],
    [48.077,88.663],
    [48.029,109.160]]

    plt.clf()
    pfid,cir,ppos,plab=plot_positions(fiducials,positions,c='y',ls='none',marker='o',fontsize=20,labels='')

    #if circle is not wanted can be removed
    cir.set_color('r')
    r = plot_rect([40,90],[50,100])
    r[0].set_linestyle(":")
    
    p = plot_poly([[45,95],[55,100],[35,105]])
    p[0].set_linestyle("-.")
    plt.draw()




import matplotlib
import matplotlib.pyplot as plt
import pdb



def demo_display(testfunc):
    """A wrapper function that makes the wrapped function work independently on environment (Ipython, Jupyter notebook, or shell). Use as a decorator on test functions to generate an output appropriate for the situation."""
    
    pass

def test_demo_display():
    pass

def get_max_size():
    """returns max size for figure, in OS independent way."""
    try:
        import psutil
        if psutil.OSX:   #True (OSX is deprecated) VC:MACOS doesn't work
            import AppKit
            screen=AppKit.NSScreen.screens()[0]  #if multiple screens
            xs,ys=screen.frame().size.width, screen.frame().size.height
        elif psutil.WINDOWS: #False
            from win32api import GetSystemMetrics
            xs,ys=GetSystemMetrics(0),GetSystemMetrics(1)  #screen size in pixel
        elif psutil.LINUX:   #False
            from Xlib import display as Xd
            resolution = Xd.Display().screen().root.get_geometry()
            xs,ys= resolution.width,resolution.height
    except:
        print ("Unrecognized OS") #not tested on different backends (only Qt)
        raise NotImplementedError
        
    return xs, ys    

def maximize(backend=None,fullscreen=False,verbose=False):
    """Maximize window independently on backend.
    
    Fullscreen sets fullscreen mode, that is same as maximized, but it doesn't have title bar (press key F to toggle full screen mode)."""
    
    #see https://stackoverflow.com/questions/12439588/how-to-maximize-a-plt-show-window-using-python

    if backend is None:
        backend=matplotlib.get_backend()
    mng = plt.get_current_fig_manager()
    if verbose:
        print ("function maximize:\nbackend: %s\nmanager:%s"%(backend,mng))
    if fullscreen:
        mng.full_screen_toggle()
    else:   
        if backend == 'wxAgg':
            mng.frame.Maximize(True)
        elif backend == 'Qt4Agg' or backend == 'Qt5Agg':
            mng.window.showMaximized()
        elif backend == 'TkAgg':
            mng.window.state('zoomed') #works fine on Windows!
        elif backend == 'module://ipykernel.pylab.backend_inline':
            #it was mng.resize(xsi,ysi) 
            #added 2020/05/26
            #works fine on Windows
            dpi = matplotlib.rcParams['figure.dpi']
            xs,ys = get_max_size()
            
            xsi,ysi = xs/dpi,ys/dpi  #convert from pixel to inches for .resize
            #xsi, ysi = xsi/2, ysi/2 #additional reduction to make fonts visible inline in a notebook
            xsi, ysi = xsi, ysi #additional reduction to make fonts visible inline in a notebook
            #matplotlib.rcParams['figure.figsize'] = [xsi,ysi] #questo funziona su WIN matplotlib inline 2020/05/26
            #mng.resize(xsi,ysi)
            params = {'legend.fontsize': 'x-large',
            'figure.figsize': (xsi,ysi),
            'axes.labelsize': 'x-large',
            'axes.titlesize':'x-large',
            'xtick.labelsize':'x-large',
            'ytick.labelsize':'x-large'}
            matplotlib.rcParams.update(params)
        else:
            xs,ys = get_max_size()
            try:
                mng.resize(xs,ys) 
            except:
                print ("Unrecognized backend: ",backend) #not tested on different backends (only Qt)
                raise NotImplementedError
                  
    #plt.show() #2020/05/26 removed to avoid showing empty grid in notebook inline.
    
    plt.pause(0.1) #this is needed to make sure following processing gets applied (e.g. tight_layout)
    #plt.draw()  #I am sure I had used the other version because this was not working in the past. I try again to try to prevent creation of blank screen in jupyter inline terminal.   
    
def list_backends():        
    """list all the available backends."""
    #from https://stackoverflow.com/questions/3285193/how-to-switch-backends-in-matplotlib-python?utm_medium=organic&utm_source=google_rich_qa&utm_campaign=google_rich_qa
    gui_env = [i for i in matplotlib.rcsetup.interactive_bk]
    non_gui_backends = matplotlib.rcsetup.non_interactive_bk
    print ("Non Gui backends are:", non_gui_backends)
    print ("Gui backends I will test for", gui_env)
    for gui in gui_env:
        print ("testing", gui)
        try:
            matplotlib.use(gui,warn=False, force=True)
            from matplotlib import pyplot as plt
            print ("    ",gui, "Is Available")
            plt.plot([1.5,2.0,2.5])
            fig = plt.gcf()
            fig.suptitle(gui)
            plt.show()
            print ("Using ..... ",matplotlib.get_backend())
        except:
            print ("    ",gui, "Not found")
    return gui_env.extend(non_gui_backends)

def test_backend(gui_env=None):
    """test a list of interactive backend."""
    # from https://stackoverflow.com/questions/3285193/how-to-switch-backends-in-matplotlib-python?utm_medium=organic&utm_source=google_rich_qa&utm_campaign=google_rich_qa
    if gui_env is None: gui_env = ['TKAgg','GTKAgg','Qt4Agg','WXAgg']
    for gui in gui_env:
        try:
            print ("testing", gui)
            matplotlib.use(gui,warn=False, force=True)
            from matplotlib import pyplot as plt
            break
        except:
            continue
    print ("Using:",matplotlib.get_backend())
    
    
def test_maximize():
    """make a few subplots and maximize window"""
    plt.figure()
    ax1=plt.subplot(121)
    plt.xlabel('X (mm)')
    plt.ylabel('Y (mm)') 
    plt.plot([1,3,6],[12,33,-1])

    ax2=plt.subplot(122,sharex=ax1,sharey=ax1)
    plt.xlabel('X (mm)')
    plt.ylabel('Y (mm)') 
    plt.plot([1,3,6],[3,32,-13])
    
    maximize()
    plt.tight_layout()
    plt.show()


from setuptools import find_packages, setup

"""with this install works only if numpy is already installed, otherwise fails on wheel."""

#p = find_packages('.')
#p = find_packages("src", exclude=["test"]),

# see https://docs.python.org/2/distutils/examples.html#pure-python-distribution-by-package
setup(
  name='pyXsurf',
  version='1.6.1',
  description="Python library for X-Ray Optics, Metrology Data Analysis and Telescopes Design.",
  url='https://github.com/vincenzooo/pyXSurf',
  author='Vincenzo Cotroneo',
  author_email='vincenzo.cotroneo@inaf.it',
  install_requires=['wheel','numpy','matplotlib','scipy','IPython','astropy'],
  #package_dir={'': 'pyxsurf'} #this makes me import as old style e.g. from pySurf import data2D, but can create overlapping, e.g. `test` or `plotting` may be used for other things.
  #package_dir={'pySurf': 'pyxsurf/pySurf',
  #             'dataIO': 'pyxsurf/dataIO'}
  package_dir={'': 'source'},
  #packages=p,
  packages = ['pySurf','dataIO','utilities','plotting','pyProfile','utilities.imaging','pySurf.readers','dataIO.config'],
  setup_requires=['numpy','astropy'],
  include_package_data=True
)




from setuptools import find_packages

# or
from setuptools import find_namespace_packages

'''
#from https://docs.pytho
n.org/3/distutils/setupscript.html

from distutils.core import setup

setup(    name='pyXsurf',
    version='0.1.1',
    packages=['pyXsurf'],
    description="Python library for X-Ray Optics, Metrology Data Analysis and Telescopes Design.",
    url='https://github.com/vincenzooo/pyXSurf',
    author='Vincenzo Cotroneo',
    author_email='vincenzo.cotroneo@inaf.it',
    #url='https://github.com/vincenzooo/pyXSurf',
    )
'''     
     
'''

from setuptools import setup
setup(
    name='pyXsurf',
    version='0.1.1',
    packages=['pyXsurf'],
    description="Python library for X-Ray Optics, Metrology Data Analysis and Telescopes Design.",
    url='https://github.com/vincenzooo/pyXSurf',
    author='Vincenzo Cotroneo',
    author_email='vincenzo.cotroneo@inaf.it',
    keywords=['surfaces', 'metrology', 'PSD'],
    tests_require=[
        'pytest'
    ],
    package_data={
        # include json and pkl files
        '': ['*.json', 'models/*.pkl', 'models/*.json'],
    },
    include_package_data=False,
    python_requires='>=3'
)

# from 

'''


import numpy as np
from math import factorial
from numpy.ma import masked_array
import scipy
from scipy.misc import factorial
import scipy.stats as stat
import pdb

#Define Zernike radial polynomials
def rnm(n,m,rho):
    """
    Return an array with the zernike Rnm polynomial calculated at rho points.
    
    
    **ARGUMENTS:**
    
        === ==========================================
        n    n order of the Zernike polynomial
        m    m order of the Zernike polynomial
        rho  Matrix containing the radial coordinates. 
        === ==========================================       
    
    .. note:: For rho>1 the returned value is 0
    
    .. note:: Values for rho<0 are silently returned as rho=0
    
    """
    
    if(type(n) is not int):
        raise Exception("n must be integer")
    if(type(m) is not int):
        raise Exception("m must be integer")
    if (n-m)%2!=0:
        raise Exception("n-m must be even")
    if abs(m)>n:
        raise Exception("The following must be true |m|<=n")
    mask=np.where(rho<=1,False,True)
    
    if(n==0 and m==0):
        return  masked_array(data=np.ones(np.shape(rho)), mask=mask)
    rho=np.where(rho<0,0,rho)
    Rnm=np.zeros(rho.shape)
    S=(n-abs(m))/2
    for s in range (0,S+1):
        CR=pow(-1,s)*factorial(n-s)/ \
            (factorial(s)*factorial(-s+(n+abs(m))/2)* \
            factorial(-s+(n-abs(m))/2))
        p=CR*pow(rho,n-2*s)
        Rnm=Rnm+p
    return masked_array(data=Rnm, mask=mask)
    
def zernike(n,m,rho,theta):
    """
    Returns the an array with the Zernike polynomial evaluated in the rho and 
    theta.
    
    **ARGUMENTS:** 
    
    ===== ==========================================     
    n     n order of the Zernike polynomial
    m     m order of the Zernike polynomial
    rho   Matrix containing the radial coordinates. 
    theta Matrix containing the angular coordinates.
    ===== ==========================================
 
    .. note:: For rho>1 the returned value is 0
    
    .. note:: Values for rho<0 are silently returned as rho=0
    """
    
    
    Rnm=rnm(n,m,rho)
    
    NC=np.sqrt(2*(n+1))
    S=(n-abs(m))/2
    
    if m>0:
        Zmn=NC*Rnm*np.cos(m*theta)
    #las funciones cos() y sin() de scipy tienen problemas cuando la grilla
    # tiene dimension cero
    
    elif m<0:
        Zmn=NC*Rnm*np.sin(m*theta)
    else:
        Zmn=np.sqrt(0.5)*NC*Rnm
    return Zmn

def zmodes(N):
    """
    Construct Zernike mode vectors in standard ordering
    Includes all modes up to radial order N
    """
    r = 0 #Starting radial mode
    radial = []
    azimuthal = []
    z = []
    while np.size(radial) < N:
        if r % 2 == 0:
            m = 0 #Set starting azimuthal mode to 0
        else:
            m = 1 #Set starting azimuthal mode to 1
        while m <= r and np.size(radial) < N:
            #Get current z number
            z = np.size(azimuthal) + 1
            #If z is odd, append sine first
            #Append negative and positive m
            if z % 2 == 1:
                azimuthal.append(-m)
            else:
                azimuthal.append(m)
            radial.append(r)
            if m > 0:
                if z % 2 == 1:
                    azimuthal.append(m)
                else:
                    azimuthal.append(-m)
                radial.append(r)
            m = m + 2
        r = r + 1 #Increment radial order
    return np.array(radial[:N],order='F').astype('int'),\
           np.array(azimuthal[:N],order='F').astype('int')

def zmatrix(rho,theta,N,r=None,m=None):
    """
    Formulate Zernike least squares fitting matrix
    Requires rho and theta vectors, normalized to rhomax=1
    """
    #Create mode vectors
    if r is None:
        r,m = zmodes(N)

    #Form matrix
    A = np.zeros((np.size(rho),np.size(r)))

    #Populate matrix columns
    for i in range(np.size(r)):
        A[:,i] = zernike(int(r[i]),int(m[i]),rho,theta)

    return A

def carttopolar(x,y,cx,cy,rad):
    #Convert to polar
    r = (sqrt((x-cx)**2+(y-cy)**2))/rad
    #Remove invalid points
    mask = r <= 1
    r = r[mask]
    x = x[mask]
    y = y[mask]
    theta = arctan2((y-cy),(x-cx))

    return r, theta, mask

#Reconstruct surface on unique x and y vectors for zernike coefficients
def zernsurf(x,y,cx,cy,rad,coeff,r=None,m=None):
    if size(unique(x)) == size(x):
        x,y = meshgrid(x,y)
    rho = sqrt((x-cx)**2+(y-cy)**2)/rad
    theta = arctan2((y-cy),(x-cx))
    heights = zeros(shape(x))

    if r is None:
        r,m = zmodes(size(coeff))

    for i in range(size(r)):
        heights = heights + coeff[i]*zernike(int(r[i]),int(m[i]),rho,theta)

    #Set outside pixels to NaN
    heights[where(rho>1.)] = NaN

    return heights.data

def fitimg(img,N=20,r=None,m=None):
    """
    Perform Zernike fit on an image.
    Zernike domain is defined over the full image array.
    """
    #Construct rho and theta vectors
    x,y = np.meshgrid(np.linspace(-1.,1.,np.shape(img)[0]),\
                      np.linspace(-1.,1.,np.shape(img)[1]))
    rho = np.sqrt(x**2+y**2)
    theta = np.arctan2(y,x)

    #Flatten and remove NaNs
    rho = rho.flatten()
    theta = theta.flatten()

    #Get b vector
    b = img.flatten()

    #Get A matrix
    A = zmatrix(rho[~np.isnan(b)],theta[~np.isnan(b)],N,r=r,m=m)

    #Solve for coefficients
    c = scipy.linalg.lstsq(A,b[~np.isnan(b)])

    #Reconstruct fit image
    coeff = c[0]
    A = zmatrix(rho,theta,N,r=r,m=m)
    fit = np.dot(A,coeff)
    fit = fit.reshape(np.shape(x))
    fit[np.isnan(img)] = np.nan

    return c,fit.reshape(np.shape(x))

def fitvec(x,y,z,N=20,r=None,m=None):
    """
    Perform Zernike fit on an image.
    Zernike domain is defined over the full image array.
    """
    #Construct rho and theta vectors
    rho = np.sqrt(x**2+y**2)
    theta = np.arctan2(y,x)
    rho = rho/rho.max()

    #Get A matrix
    A = zmatrix(rho,theta,N,r=r,m=m)

    #Solve for coefficients
    c = scipy.linalg.lstsq(A,z)

    return c
    
    

#Function to output Zernike coefficients and RMS fit error for x,y,z txt file
def zcoeff(filename,save=False,cx=0.,cy=0.,rad=1.,order=20,r=None,m=None,**kwags):
    #Read in surface data
    if type(filename) is str:
        d = genfromtxt(filename,**kwags)
    else:
        d = filename

    if shape(d)[0]==3:
        sagx, sagy, sagz = d[0],d[1],d[2]
    else:
##        #Strip NaN rows/columns
##        while sum(isnan(d[0]))==shape(d)[1]:
##            d = d[1:]
##        while sum(isnan(d[-1]))==shape(d)[1]:
##            d = d[:-1]
##        newsize = shape(d)[0]
##        while sum(isnan(d[:,0]))==newsize:
##            d = d[:,1:]
##        while sum(isnan(d[:,-1]))==newsize:
##            d = d[:,:-1]
        x,y=meshgrid(arange(shape(d)[0],dtype='float'),\
                     arange(shape(d)[1],dtype='float'))
##        ind = invert(isnan(d))
##        d2 = d[ind]
##        x = x[ind]
##        y = y[ind]
        x = x.flatten()
        y = y.flatten()
        d2 = d.flatten()
        sagx = []
        sagy = []
        sagz = []
        for i in range(size(d2)):
            if invert(isnan(d2[i])):
                sagx.append(x[i])
                sagy.append(y[i])
                sagz.append(d2[i])
        sagx = array(sagx)
        sagy = array(sagy)
        sagz = array(sagz)

    #Convert to normalized polar coordinates for Zernike fitting
    rho, theta, mask = carttopolar(sagx,sagy,cx,cy,rad)

    #Create Zernike polynomial matrix for least squares fit
    #Using all Zernike polynomials up to radial order 20
    pdb.set_trace()
    A = zmatrix(rho,theta,order,r=r,m=m)

    #Perform least squares fit
    #0th element is the coefficient matrix
    #1st element is sum of squared residuals
    #2nd element is rank of matrix A
    #3rd element is singular values of A
    fit = scipy.linalg.lstsq(A,sagz[mask])

    #Compute fitted surface from Zernike coefficients
    if shape(d)[0] != 3:
        y,x=meshgrid(arange(shape(d)[0],dtype='float'),\
                         arange(shape(d)[1],dtype='float'))
    else:
        x,y = d[0],d[1]
    fitsurf = zernsurf(x.flatten(),y.flatten(),cx,cy,rad,fit[0],r=r,m=m)
##
##    #Do residuals match up with those from fit?
##    print sum((fitsurf-sagz)**2)
##    print fit[1]

    rms = sqrt(fit[1]/size(sagz))

    #If save=True, save coefficients to a txt file
    #First line is number of coefficients
    if save==True:
        savetxt(filename.split('.')[0]+'Coeff.txt'\
                ,insert(fit[0],0,size(fit[0])))

    return fit[0],fit[1],rms,fitsurf



from ast import literal_eval
import os
import io

try:
    from configparser import ConfigParser
except ImportError:
    from ConfigParser import ConfigParser  # ver. < 3.0


def save_section(sec,outfile):
    
    
    """save a specific section (or list of sections) to a config file."""
    
    
    try:
        sec.name
        sec=[sec]
    except AttributeError:
        assert type(sec)==list
        
    outfolder=os.path.basename(os.path.dirname(outfile))
    with open(outfile,'w') as cfgfile:
        newconf= ConfigParser()
        for s in sec:
            newconf.add_section(s.name)
            for k,v in list(s.items()):
                newconf.set(s.name,k,v)
        newconf.write(cfgfile)
        cfgfile.close()
            
def make_config(settingsFile):
    """create a configuration object from a .ini file, set defaults.
    return config and list of sections to process as specified in 'process' section
    (default to all sections)."""
    # on configparser https://docs.python.org/3/library/configparser.html
    config = ConfigParser()
    """
    these have been included in settings file.
    config.set('DEFAULT' ,'rect1', "")
    config.set('DEFAULT' ,'mfile', "")
    config.set('DEFAULT' ,'mscale', "[1.,1.,1.]")
    config.set('DEFAULT' ,'gscale', "[1.,1.,1.]")
    #config.set('DEFAULT','cane',0)  #set default
    """
    # parse existing file
    config.read(settingsFile)    

    #includelist=json.loads(config.get('process', 'includelist')) #list of settings to process
    if config.has_section('process'):
        includelist=literal_eval(config.get('process', 'includelist')) #list of settings to process
    else:
        includelist=config.sections()
    return config,includelist

def string_to_config(header):
    """ convert a string to a config object. """
    config = ConfigParser()
    buf = io.StringIO("\n".join(header))
    config.read_file(buf)
    return config